import logging

from dbnd._core.errors.base import DatabandApiError
from dbnd._core.utils.basics.text_banner import TextBanner, safe_tabulate
from dbnd._vendor import click
from dbnd.api.airflow_sync import (
    archive_airflow_instance,
    create_airflow_instance,
    list_synced_airflow_instances,
    unarchive_airflow_instance,
)


logger = logging.getLogger(__name__)


@click.group()
@click.pass_context
def airflow_sync(ctx):
    """Manage synced Airflow instances"""
    pass


@airflow_sync.command("list")
def list_airflow_instances():
    try:
        instances = list_synced_airflow_instances()
    except (LookupError, DatabandApiError) as e:
        logger.error(e)
    else:
        print_table("Synced Airflow instances", instances)


def print_table(header, instances_list):
    banner = TextBanner(header)
    banner.write(build_instances_table(instances_list))
    logger.info(banner.get_banner_str())


def build_instances_table(instances_data):
    extract_keys = (
        "is_sync_enabled",
        "base_url",
        "external_url",
        "api_mode",
    )
    headers = (
        "Active",
        "Base url",
        "External url",
        "Api Mode",
        "Instance Type",
    )
    table_data = []
    for server_info in instances_data:
        instance_row = [getattr(server_info, key, "") for key in extract_keys]
        instance_row.append(
            "Airflow" if getattr(server_info, "fetcher") == "web" else "GC Composer"
        )
        table_data.append(instance_row)
    return safe_tabulate(table_data, headers)


def validate_composer_id(ctx, param, value):
    if ctx.params["instance_type"] == "composer":
        if value is None:
            raise click.MissingParameter(
                "Client id is required for Google Cloud Composer instance", ctx, param
            )
    else:
        if value is not None:
            raise click.BadParameter(
                "Setting client id is allowed only for Google Cloud Composer instances",
                ctx,
                param,
            )
    return value


@airflow_sync.command()
@click.option(
    "--url", "-u", help="base url for instance", type=click.STRING, required=True
)
@click.option(
    "--external-url", "-e", help="external url for instance", type=click.STRING
)
@click.option(
    "--instance-type",
    "-i",
    help="type of an instance: Airflow or Google Cloud Composer",
    type=click.Choice(["airflow", "composer"], case_sensitive=False),
    required=True,
)
@click.option(
    "--api-mode",
    "-a",
    help="Airflow API connection mode",
    type=click.Choice(["rbac", "flask-admin", "experimental"]),
    default="rbac",
)
@click.option(
    "--composer-client-id",
    "-c",
    help="client id for Google Cloud Composer connection",
    type=click.STRING,
    callback=validate_composer_id,
)
def add(url, external_url, instance_type, api_mode, composer_client_id):
    fetcher = "web" if instance_type == "airflow" else instance_type
    try:
        create_airflow_instance(
            url, external_url, fetcher, api_mode, composer_client_id
        )
    except DatabandApiError as e:
        logger.error("failed with - {}".format(e.response))
    else:
        logger.info("Starting syncing instance %s", url)


@airflow_sync.command()
@click.option(
    "--url",
    "-u",
    help="base url of an instance to archive",
    type=click.STRING,
    required=True,
)
def archive(url):
    try:
        archive_airflow_instance(url)
    except DatabandApiError as e:
        logger.error("failed with - {}".format(e.response))
    else:
        logger.info("Archived instance %s", url)


@airflow_sync.command()
@click.option(
    "--url",
    "-u",
    help="base url of an instance to unarchive",
    type=click.STRING,
    required=True,
)
def unarchive(url):
    try:
        unarchive_airflow_instance(url)
    except DatabandApiError as e:
        logger.error("failed with - {}".format(e.response))
    else:
        logger.info("Unarchived instance %s", url)



import datetime

from datetime import timedelta
from typing import Iterable


def period_dates(target_date, period, step=timedelta(days=1)):
    # type: (datetime.date, datetime.timedelta, datetime.timedelta) -> Iterable[datetime.date]
    while period.total_seconds() > 0:
        yield target_date
        period -= step
        target_date -= step



from dbnd._vendor.croniter import croniter
from dbnd._vendor.marshmallow import Schema, fields


class AlertEventSchema(Schema):
    name = fields.Str()
    description = fields.Str()
    run_uid = fields.Str()


class ScheduledJobSchemaV2(Schema):
    class Meta:
        strict = True

    uid = fields.Str(attribute="DbndScheduledJob.uid", allow_none=True)
    name = fields.Str(attribute="DbndScheduledJob.name", required=True)
    cmd = fields.Str(attribute="DbndScheduledJob.cmd", required=True)
    schedule_interval = fields.Str(
        attribute="DbndScheduledJob.schedule_interval", required=True
    )
    start_date = fields.DateTime(
        allow_none=True, attribute="DbndScheduledJob.start_date", format="iso"
    )
    end_date = fields.DateTime(
        allow_none=True, attribute="DbndScheduledJob.end_date", format="iso"
    )
    readable_schedule_interval = fields.Str(
        attribute="DbndScheduledJob.readable_schedule_interval", allow_none=True
    )
    scheduled_interval_in_seconds = fields.Integer(
        attribute="DbndScheduledJob.scheduled_interval_in_seconds", allow_none=True
    )
    catchup = fields.Boolean(allow_none=True, attribute="DbndScheduledJob.catchup")
    depends_on_past = fields.Boolean(
        allow_none=True, attribute="DbndScheduledJob.depends_on_past"
    )
    retries = fields.Int(allow_none=True, attribute="DbndScheduledJob.retries")

    active = fields.Boolean(allow_none=True, attribute="DbndScheduledJob.active")
    create_user = fields.Str(allow_none=True, attribute="DbndScheduledJob.create_user")
    create_time = fields.DateTime(
        allow_none=True, attribute="DbndScheduledJob.create_time"
    )
    update_user = fields.Str(allow_none=True, attribute="DbndScheduledJob.update_user")
    update_time = fields.DateTime(
        allow_none=True, attribute="DbndScheduledJob.update_time"
    )
    from_file = fields.Boolean(allow_none=True, attribute="DbndScheduledJob.from_file")
    deleted_from_file = fields.Boolean(
        allow_none=True, attribute="DbndScheduledJob.deleted_from_file"
    )
    next_job_date = fields.DateTime(
        attribute="DbndScheduledJob.next_job_date", allow_none=True
    )
    alerts = fields.List(
        fields.Nested(AlertEventSchema),
        attribute="DbndScheduledJob.alerts",
        allow_none=True,
    )

    job_name = fields.Str(dump_only=True, attribute="DbndScheduledJob.job_name")
    job_id = fields.Str()
    last_run_uid = fields.UUID(dump_only=True)
    last_run_job = fields.Str(dump_only=True)
    last_job_date = fields.DateTime(dump_only=True)
    last_run_state = fields.Str(dump_only=True)
    latest_run_env = fields.Str(dump_only=True)
    root_task_run_uid = fields.Str(dump_only=True)
    is_airflow_synced = fields.Bool(dump_only=True)
    airflow_instance_name = fields.Str(allow_none=True)
    list_order = fields.Integer(
        attribute="DbndScheduledJob.list_order", allow_none=True
    )
    validation_errors = fields.Str(
        allow_none=True, attribute="DbndScheduledJob.validation_errors"
    )


SCHEDULE_INTERVAL_PRESETS = {
    "@once": None,
    "@hourly": "0 * * * *",
    "@daily": "0 0 * * *",
    "@weekly": "0 0 * * 0",
    "@monthly": "0 0 1 * *",
    "@yearly": "0 0 1 1 *",
}


def validate_cron(expression):
    if expression in SCHEDULE_INTERVAL_PRESETS:
        return None

    try:
        croniter.expand(expression)
    except Exception as e:
        return str(e)
    else:
        return None



import logging

from dbnd._core.decorator.decorated_task import (
    DecoratedPipelineTask,
    DecoratedPythonTask,
)
from dbnd._core.decorator.task_cls_builder import _task_decorator
from dbnd._core.parameter.parameter_builder import parameter


logger = logging.getLogger(__name__)
_default_output = parameter.output.pickle[object]


def task(*args, **kwargs):
    kwargs.setdefault("_task_type", DecoratedPythonTask)
    kwargs.setdefault("_task_default_result", _default_output)
    return _task_decorator(*args, **kwargs)


def pipeline(*args, **kwargs):
    kwargs.setdefault("_task_type", DecoratedPipelineTask)
    kwargs.setdefault("_task_default_result", parameter.output)
    return _task_decorator(*args, **kwargs)


band = pipeline
data_source_pipeline = pipeline



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: airflow_monitor.proto
"""Generated protocol buffer code."""
from google.protobuf import (
    descriptor as _descriptor,
    message as _message,
    reflection as _reflection,
    symbol_database as _symbol_database,
    timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2,
)


# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name="airflow_monitor.proto",
    package="dbnd.services.airflow_monitor",
    syntax="proto3",
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n\x15\x61irflow_monitor.proto\x12\x1d\x64\x62nd.services.airflow_monitor\x1a\x1fgoogle/protobuf/timestamp.proto"\x99\x07\n\x11\x41irflowServerInfo\x12\x10\n\x08\x62\x61se_url\x18\x01 \x01(\t\x12\x1c\n\x0f\x61irflow_version\x18\x02 \x01(\tH\x00\x88\x01\x01\x12#\n\x16\x61irflow_export_version\x18\x03 \x01(\tH\x01\x88\x01\x01\x12$\n\x17\x61irflow_monitor_version\x18\x04 \x01(\tH\x02\x88\x01\x01\x12\x16\n\tdags_path\x18\x05 \x01(\tH\x03\x88\x01\x01\x12\x16\n\tlogs_path\x18\x06 \x01(\tH\x04\x88\x01\x01\x12\x37\n\x0elast_sync_time\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.TimestampH\x05\x88\x01\x01\x12\x1b\n\x0emonitor_status\x18\x08 \x01(\tH\x06\x88\x01\x01\x12"\n\x15monitor_error_message\x18\t \x01(\tH\x07\x88\x01\x01\x12;\n\x12monitor_start_time\x18\n \x01(\x0b\x32\x1a.google.protobuf.TimestampH\x08\x88\x01\x01\x12\x34\n\x0bsynced_from\x18\x0b \x01(\x0b\x32\x1a.google.protobuf.TimestampH\t\x88\x01\x01\x12\x32\n\tsynced_to\x18\x0c \x01(\x0b\x32\x1a.google.protobuf.TimestampH\n\x88\x01\x01\x12\x15\n\x08\x61pi_mode\x18\r \x01(\tH\x0b\x88\x01\x01\x12\x1a\n\rsync_interval\x18\x0e \x01(\rH\x0c\x88\x01\x01\x12\x1c\n\x0fis_sync_enabled\x18\x0f \x01(\x08H\r\x88\x01\x01\x12\x14\n\x07\x66\x65tcher\x18\x10 \x01(\tH\x0e\x88\x01\x01\x12\x1f\n\x12\x63omposer_client_id\x18\x11 \x01(\tH\x0f\x88\x01\x01\x42\x12\n\x10_airflow_versionB\x19\n\x17_airflow_export_versionB\x1a\n\x18_airflow_monitor_versionB\x0c\n\n_dags_pathB\x0c\n\n_logs_pathB\x11\n\x0f_last_sync_timeB\x11\n\x0f_monitor_statusB\x18\n\x16_monitor_error_messageB\x15\n\x13_monitor_start_timeB\x0e\n\x0c_synced_fromB\x0c\n\n_synced_toB\x0b\n\t_api_modeB\x10\n\x0e_sync_intervalB\x12\n\x10_is_sync_enabledB\n\n\x08_fetcherB\x15\n\x13_composer_client_id"\x83\x01\n\x1cPostAirflowServerInfoRequest\x12M\n\x13\x61irflow_server_info\x18\x01 \x01(\x0b\x32\x30.dbnd.services.airflow_monitor.AirflowServerInfo\x12\x14\n\x0c\x65xternal_url\x18\x02 \x01(\t"\x1f\n\x1dPostAirflowServerInfoResponse2\xac\x01\n\x15\x41irflowMonitorService\x12\x92\x01\n\x15PostAirflowServerInfo\x12;.dbnd.services.airflow_monitor.PostAirflowServerInfoRequest\x1a<.dbnd.services.airflow_monitor.PostAirflowServerInfoResponseb\x06proto3',
    dependencies=[google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,],
)


_AIRFLOWSERVERINFO = _descriptor.Descriptor(
    name="AirflowServerInfo",
    full_name="dbnd.services.airflow_monitor.AirflowServerInfo",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="base_url",
            full_name="dbnd.services.airflow_monitor.AirflowServerInfo.base_url",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="airflow_version",
            full_name="dbnd.services.airflow_monitor.AirflowServerInfo.airflow_version",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="airflow_export_version",
            full_name="dbnd.services.airflow_monitor.AirflowServerInfo.airflow_export_version",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="airflow_monitor_version",
            full_name="dbnd.services.airflow_monitor.AirflowServerInfo.airflow_monitor_version",
            index=3,
            number=4,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="dags_path",
            full_name="dbnd.services.airflow_monitor.AirflowServerInfo.dags_path",
            index=4,
            number=5,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="logs_path",
            full_name="dbnd.services.airflow_monitor.AirflowServerInfo.logs_path",
            index=5,
            number=6,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="last_sync_time",
            full_name="dbnd.services.airflow_monitor.AirflowServerInfo.last_sync_time",
            index=6,
            number=7,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="monitor_status",
            full_name="dbnd.services.airflow_monitor.AirflowServerInfo.monitor_status",
            index=7,
            number=8,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="monitor_error_message",
            full_name="dbnd.services.airflow_monitor.AirflowServerInfo.monitor_error_message",
            index=8,
            number=9,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="monitor_start_time",
            full_name="dbnd.services.airflow_monitor.AirflowServerInfo.monitor_start_time",
            index=9,
            number=10,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="synced_from",
            full_name="dbnd.services.airflow_monitor.AirflowServerInfo.synced_from",
            index=10,
            number=11,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="synced_to",
            full_name="dbnd.services.airflow_monitor.AirflowServerInfo.synced_to",
            index=11,
            number=12,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="api_mode",
            full_name="dbnd.services.airflow_monitor.AirflowServerInfo.api_mode",
            index=12,
            number=13,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="sync_interval",
            full_name="dbnd.services.airflow_monitor.AirflowServerInfo.sync_interval",
            index=13,
            number=14,
            type=13,
            cpp_type=3,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="is_sync_enabled",
            full_name="dbnd.services.airflow_monitor.AirflowServerInfo.is_sync_enabled",
            index=14,
            number=15,
            type=8,
            cpp_type=7,
            label=1,
            has_default_value=False,
            default_value=False,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="fetcher",
            full_name="dbnd.services.airflow_monitor.AirflowServerInfo.fetcher",
            index=15,
            number=16,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="composer_client_id",
            full_name="dbnd.services.airflow_monitor.AirflowServerInfo.composer_client_id",
            index=16,
            number=17,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name="_airflow_version",
            full_name="dbnd.services.airflow_monitor.AirflowServerInfo._airflow_version",
            index=0,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
        _descriptor.OneofDescriptor(
            name="_airflow_export_version",
            full_name="dbnd.services.airflow_monitor.AirflowServerInfo._airflow_export_version",
            index=1,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
        _descriptor.OneofDescriptor(
            name="_airflow_monitor_version",
            full_name="dbnd.services.airflow_monitor.AirflowServerInfo._airflow_monitor_version",
            index=2,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
        _descriptor.OneofDescriptor(
            name="_dags_path",
            full_name="dbnd.services.airflow_monitor.AirflowServerInfo._dags_path",
            index=3,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
        _descriptor.OneofDescriptor(
            name="_logs_path",
            full_name="dbnd.services.airflow_monitor.AirflowServerInfo._logs_path",
            index=4,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
        _descriptor.OneofDescriptor(
            name="_last_sync_time",
            full_name="dbnd.services.airflow_monitor.AirflowServerInfo._last_sync_time",
            index=5,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
        _descriptor.OneofDescriptor(
            name="_monitor_status",
            full_name="dbnd.services.airflow_monitor.AirflowServerInfo._monitor_status",
            index=6,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
        _descriptor.OneofDescriptor(
            name="_monitor_error_message",
            full_name="dbnd.services.airflow_monitor.AirflowServerInfo._monitor_error_message",
            index=7,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
        _descriptor.OneofDescriptor(
            name="_monitor_start_time",
            full_name="dbnd.services.airflow_monitor.AirflowServerInfo._monitor_start_time",
            index=8,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
        _descriptor.OneofDescriptor(
            name="_synced_from",
            full_name="dbnd.services.airflow_monitor.AirflowServerInfo._synced_from",
            index=9,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
        _descriptor.OneofDescriptor(
            name="_synced_to",
            full_name="dbnd.services.airflow_monitor.AirflowServerInfo._synced_to",
            index=10,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
        _descriptor.OneofDescriptor(
            name="_api_mode",
            full_name="dbnd.services.airflow_monitor.AirflowServerInfo._api_mode",
            index=11,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
        _descriptor.OneofDescriptor(
            name="_sync_interval",
            full_name="dbnd.services.airflow_monitor.AirflowServerInfo._sync_interval",
            index=12,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
        _descriptor.OneofDescriptor(
            name="_is_sync_enabled",
            full_name="dbnd.services.airflow_monitor.AirflowServerInfo._is_sync_enabled",
            index=13,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
        _descriptor.OneofDescriptor(
            name="_fetcher",
            full_name="dbnd.services.airflow_monitor.AirflowServerInfo._fetcher",
            index=14,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
        _descriptor.OneofDescriptor(
            name="_composer_client_id",
            full_name="dbnd.services.airflow_monitor.AirflowServerInfo._composer_client_id",
            index=15,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
    ],
    serialized_start=90,
    serialized_end=1011,
)


_POSTAIRFLOWSERVERINFOREQUEST = _descriptor.Descriptor(
    name="PostAirflowServerInfoRequest",
    full_name="dbnd.services.airflow_monitor.PostAirflowServerInfoRequest",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="airflow_server_info",
            full_name="dbnd.services.airflow_monitor.PostAirflowServerInfoRequest.airflow_server_info",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="external_url",
            full_name="dbnd.services.airflow_monitor.PostAirflowServerInfoRequest.external_url",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1014,
    serialized_end=1145,
)


_POSTAIRFLOWSERVERINFORESPONSE = _descriptor.Descriptor(
    name="PostAirflowServerInfoResponse",
    full_name="dbnd.services.airflow_monitor.PostAirflowServerInfoResponse",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=1147,
    serialized_end=1178,
)

_AIRFLOWSERVERINFO.fields_by_name[
    "last_sync_time"
].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_AIRFLOWSERVERINFO.fields_by_name[
    "monitor_start_time"
].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_AIRFLOWSERVERINFO.fields_by_name[
    "synced_from"
].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_AIRFLOWSERVERINFO.fields_by_name[
    "synced_to"
].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_AIRFLOWSERVERINFO.oneofs_by_name["_airflow_version"].fields.append(
    _AIRFLOWSERVERINFO.fields_by_name["airflow_version"]
)
_AIRFLOWSERVERINFO.fields_by_name[
    "airflow_version"
].containing_oneof = _AIRFLOWSERVERINFO.oneofs_by_name["_airflow_version"]
_AIRFLOWSERVERINFO.oneofs_by_name["_airflow_export_version"].fields.append(
    _AIRFLOWSERVERINFO.fields_by_name["airflow_export_version"]
)
_AIRFLOWSERVERINFO.fields_by_name[
    "airflow_export_version"
].containing_oneof = _AIRFLOWSERVERINFO.oneofs_by_name["_airflow_export_version"]
_AIRFLOWSERVERINFO.oneofs_by_name["_airflow_monitor_version"].fields.append(
    _AIRFLOWSERVERINFO.fields_by_name["airflow_monitor_version"]
)
_AIRFLOWSERVERINFO.fields_by_name[
    "airflow_monitor_version"
].containing_oneof = _AIRFLOWSERVERINFO.oneofs_by_name["_airflow_monitor_version"]
_AIRFLOWSERVERINFO.oneofs_by_name["_dags_path"].fields.append(
    _AIRFLOWSERVERINFO.fields_by_name["dags_path"]
)
_AIRFLOWSERVERINFO.fields_by_name[
    "dags_path"
].containing_oneof = _AIRFLOWSERVERINFO.oneofs_by_name["_dags_path"]
_AIRFLOWSERVERINFO.oneofs_by_name["_logs_path"].fields.append(
    _AIRFLOWSERVERINFO.fields_by_name["logs_path"]
)
_AIRFLOWSERVERINFO.fields_by_name[
    "logs_path"
].containing_oneof = _AIRFLOWSERVERINFO.oneofs_by_name["_logs_path"]
_AIRFLOWSERVERINFO.oneofs_by_name["_last_sync_time"].fields.append(
    _AIRFLOWSERVERINFO.fields_by_name["last_sync_time"]
)
_AIRFLOWSERVERINFO.fields_by_name[
    "last_sync_time"
].containing_oneof = _AIRFLOWSERVERINFO.oneofs_by_name["_last_sync_time"]
_AIRFLOWSERVERINFO.oneofs_by_name["_monitor_status"].fields.append(
    _AIRFLOWSERVERINFO.fields_by_name["monitor_status"]
)
_AIRFLOWSERVERINFO.fields_by_name[
    "monitor_status"
].containing_oneof = _AIRFLOWSERVERINFO.oneofs_by_name["_monitor_status"]
_AIRFLOWSERVERINFO.oneofs_by_name["_monitor_error_message"].fields.append(
    _AIRFLOWSERVERINFO.fields_by_name["monitor_error_message"]
)
_AIRFLOWSERVERINFO.fields_by_name[
    "monitor_error_message"
].containing_oneof = _AIRFLOWSERVERINFO.oneofs_by_name["_monitor_error_message"]
_AIRFLOWSERVERINFO.oneofs_by_name["_monitor_start_time"].fields.append(
    _AIRFLOWSERVERINFO.fields_by_name["monitor_start_time"]
)
_AIRFLOWSERVERINFO.fields_by_name[
    "monitor_start_time"
].containing_oneof = _AIRFLOWSERVERINFO.oneofs_by_name["_monitor_start_time"]
_AIRFLOWSERVERINFO.oneofs_by_name["_synced_from"].fields.append(
    _AIRFLOWSERVERINFO.fields_by_name["synced_from"]
)
_AIRFLOWSERVERINFO.fields_by_name[
    "synced_from"
].containing_oneof = _AIRFLOWSERVERINFO.oneofs_by_name["_synced_from"]
_AIRFLOWSERVERINFO.oneofs_by_name["_synced_to"].fields.append(
    _AIRFLOWSERVERINFO.fields_by_name["synced_to"]
)
_AIRFLOWSERVERINFO.fields_by_name[
    "synced_to"
].containing_oneof = _AIRFLOWSERVERINFO.oneofs_by_name["_synced_to"]
_AIRFLOWSERVERINFO.oneofs_by_name["_api_mode"].fields.append(
    _AIRFLOWSERVERINFO.fields_by_name["api_mode"]
)
_AIRFLOWSERVERINFO.fields_by_name[
    "api_mode"
].containing_oneof = _AIRFLOWSERVERINFO.oneofs_by_name["_api_mode"]
_AIRFLOWSERVERINFO.oneofs_by_name["_sync_interval"].fields.append(
    _AIRFLOWSERVERINFO.fields_by_name["sync_interval"]
)
_AIRFLOWSERVERINFO.fields_by_name[
    "sync_interval"
].containing_oneof = _AIRFLOWSERVERINFO.oneofs_by_name["_sync_interval"]
_AIRFLOWSERVERINFO.oneofs_by_name["_is_sync_enabled"].fields.append(
    _AIRFLOWSERVERINFO.fields_by_name["is_sync_enabled"]
)
_AIRFLOWSERVERINFO.fields_by_name[
    "is_sync_enabled"
].containing_oneof = _AIRFLOWSERVERINFO.oneofs_by_name["_is_sync_enabled"]
_AIRFLOWSERVERINFO.oneofs_by_name["_fetcher"].fields.append(
    _AIRFLOWSERVERINFO.fields_by_name["fetcher"]
)
_AIRFLOWSERVERINFO.fields_by_name[
    "fetcher"
].containing_oneof = _AIRFLOWSERVERINFO.oneofs_by_name["_fetcher"]
_AIRFLOWSERVERINFO.oneofs_by_name["_composer_client_id"].fields.append(
    _AIRFLOWSERVERINFO.fields_by_name["composer_client_id"]
)
_AIRFLOWSERVERINFO.fields_by_name[
    "composer_client_id"
].containing_oneof = _AIRFLOWSERVERINFO.oneofs_by_name["_composer_client_id"]
_POSTAIRFLOWSERVERINFOREQUEST.fields_by_name[
    "airflow_server_info"
].message_type = _AIRFLOWSERVERINFO
DESCRIPTOR.message_types_by_name["AirflowServerInfo"] = _AIRFLOWSERVERINFO
DESCRIPTOR.message_types_by_name[
    "PostAirflowServerInfoRequest"
] = _POSTAIRFLOWSERVERINFOREQUEST
DESCRIPTOR.message_types_by_name[
    "PostAirflowServerInfoResponse"
] = _POSTAIRFLOWSERVERINFORESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

AirflowServerInfo = _reflection.GeneratedProtocolMessageType(
    "AirflowServerInfo",
    (_message.Message,),
    {
        "DESCRIPTOR": _AIRFLOWSERVERINFO,
        "__module__": "airflow_monitor_pb2"
        # @@protoc_insertion_point(class_scope:dbnd.services.airflow_monitor.AirflowServerInfo)
    },
)
_sym_db.RegisterMessage(AirflowServerInfo)

PostAirflowServerInfoRequest = _reflection.GeneratedProtocolMessageType(
    "PostAirflowServerInfoRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _POSTAIRFLOWSERVERINFOREQUEST,
        "__module__": "airflow_monitor_pb2"
        # @@protoc_insertion_point(class_scope:dbnd.services.airflow_monitor.PostAirflowServerInfoRequest)
    },
)
_sym_db.RegisterMessage(PostAirflowServerInfoRequest)

PostAirflowServerInfoResponse = _reflection.GeneratedProtocolMessageType(
    "PostAirflowServerInfoResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _POSTAIRFLOWSERVERINFORESPONSE,
        "__module__": "airflow_monitor_pb2"
        # @@protoc_insertion_point(class_scope:dbnd.services.airflow_monitor.PostAirflowServerInfoResponse)
    },
)
_sym_db.RegisterMessage(PostAirflowServerInfoResponse)


_AIRFLOWMONITORSERVICE = _descriptor.ServiceDescriptor(
    name="AirflowMonitorService",
    full_name="dbnd.services.airflow_monitor.AirflowMonitorService",
    file=DESCRIPTOR,
    index=0,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_start=1181,
    serialized_end=1353,
    methods=[
        _descriptor.MethodDescriptor(
            name="PostAirflowServerInfo",
            full_name="dbnd.services.airflow_monitor.AirflowMonitorService.PostAirflowServerInfo",
            index=0,
            containing_service=None,
            input_type=_POSTAIRFLOWSERVERINFOREQUEST,
            output_type=_POSTAIRFLOWSERVERINFORESPONSE,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
    ],
)
_sym_db.RegisterServiceDescriptor(_AIRFLOWMONITORSERVICE)

DESCRIPTOR.services_by_name["AirflowMonitorService"] = _AIRFLOWMONITORSERVICE

# @@protoc_insertion_point(module_scope)



import logging
import os
import shlex
import signal
import subprocess

from subprocess import list2cmdline
from typing import Any, Dict, List, Optional

import six

from dbnd import parameter, task
from dbnd._core.commands import log_metric
from dbnd._core.current import try_get_current_task_run
from dbnd._core.errors import DatabandConfigError
from dbnd._core.errors.friendly_error.task_execution import failed_to_run_cmd
from dbnd._core.run.databand_run import DatabandRun
from dbnd._core.utils.basics.signal_utils import safe_signal
from dbnd._core.utils.platform import windows_compatible_mode


logger = logging.getLogger(__name__)


@task(popen_kwargs=parameter(empty_default=True)[Dict[str, str]])
def bash_cmd(
    cmd=None,
    args=None,
    check_retcode=0,
    cwd=None,
    env=None,
    dbnd_env=True,
    output_encoding="utf-8",
    popen_kwargs=None,
    wait_for_termination_s=5,
    shell=False,
):
    # type:( str, List[str], Optional[int], str, Dict[str,str], bool, str, Dict[str,Any]) -> int
    if popen_kwargs is None:
        popen_kwargs = dict()
    popen_kwargs = popen_kwargs.copy()

    if cmd and args:
        raise DatabandConfigError("You should not provide cmd and args ")

    if cmd:
        if shell:
            args = cmd
        else:
            args = shlex.split(cmd)
    elif args:
        args = list(map(str, args))
        cmd = list2cmdline(args)
        if shell:
            args = cmd

    logger.info("Running: " + cmd)  # To simplify rerunning failing tests

    if dbnd_env and DatabandRun.has_instance():
        env = env or os.environ.copy()
        dbnd_env_vars = DatabandRun.get_instance().get_context_spawn_env()
        logger.info(
            "Exporting the following env vars:\n%s",
            "\n".join(["{}={}".format(k, v) for k, v in dbnd_env_vars.items()]),
        )
        env.update()

    def preexec_fn():
        if windows_compatible_mode:
            return
        # Restore default signal disposition and invoke setsid
        for sig in ("SIGPIPE", "SIGXFZ", "SIGXFSZ"):
            if hasattr(signal, sig):
                safe_signal(getattr(signal, sig), signal.SIG_DFL)
        os.setsid()

    process = subprocess.Popen(
        args,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=-1,
        universal_newlines=True,
        env=env,
        preexec_fn=preexec_fn,
        cwd=cwd,
        shell=shell,
        **popen_kwargs
    )

    try:
        task_run = try_get_current_task_run()
        if task_run:
            task_run.task.process = process

        logger.info("Process is running, output:")
        # While command is running let read it's output
        output = []
        while True:
            line = process.stdout.readline()
            if line == "" or line == b"":
                break
            line = safe_decode(line, output_encoding).rstrip()

            logger.info("out: %s", line)
            # keep last 1000 lines only
            output.append(line)
            if len(output) > 1500:
                output = output[-1000:]

        returncode = process.wait()
        logger.info("Command exited with return code %s", process.returncode)
        if check_retcode is not None and returncode != check_retcode:
            raise failed_to_run_cmd(
                "Bash command failed", cmd_str=cmd, return_code=returncode
            )
        return returncode
    except Exception:
        logger.info("Received interrupt. Terminating subprocess and waiting")
        try:
            process.terminate()
            process.wait(wait_for_termination_s)
        except Exception:
            pass
        raise


def safe_decode(line, output_encoding):
    if isinstance(line, six.string_types):
        return line
    return line.decode(output_encoding)


@task(popen_kwargs=parameter(empty_default=True)[Dict[str, str]])
def bash_script(
    script=None,
    check_retcode=0,
    cwd=None,
    env=None,
    dbnd_env=True,
    output_encoding="utf-8",
    popen_kwargs=None,
):
    # type:( str, Optional[int],str, Dict[str,str], bool, str, Dict[str,Any]) -> int

    # we need a working folder to create bash script
    task_run = try_get_current_task_run()
    if task_run:
        script_dir = str(task_run.local_task_run_root)
    else:
        script_dir = None

    bash_script_path = os.path.join(script_dir, "bash_cmd.sh")
    with open(bash_script_path, "wb") as bs:
        bs.write(bytes(script, "utf_8"))

    log_metric("bash_script", bash_script_path)

    logger.info("Bash script location: %s", bash_script_path)
    args = ["bash", bash_script_path]
    return bash_cmd.func(
        args=args,
        check_retcode=check_retcode,
        cwd=cwd or script_dir,
        env=env,
        dbnd_env=dbnd_env,
        output_encoding=output_encoding,
        popen_kwargs=popen_kwargs,
    )



import datetime

import pytz

from dbnd._core.utils.timezone import utcnow
from dbnd._vendor import pendulum
from targets.values import ValueType


class DateAlias(object):
    now = "now"
    today = "today"
    yesterday = "yesterday"


class DateValueType(ValueType):
    """
       ValueType whose value is a :py:class:`~datetime.date`.

       A DateValueType is a Date string formatted ``YYYY-MM-DD``. For example, ``2013-07-10`` specifies
       July 10, 2013.

       DateValueTypes are 90% of the time used to be interpolated into file system paths or the like.
       Here is a gentle reminder of how to interpolate date parameters into strings:

       .. code:: python

           class MyTask(dbnd.Task):
               date = databand.DateValueType()

               def run(self):
                   templated_path = "/my/path/to/my/dataset/{date:%Y/%m/%d}/"
                   instantiated_path = templated_path.format(date=self.date)
                   # print(instantiated_path) --> /my/path/to/my/dataset/2016/06/09/
                   # ... use instantiated_path ...

       To set this parameter to default to the current day. You can write code like this:

       .. code:: python

           import datetime

           class MyTask(dbnd.Task):
               date = databand.DateValueType(default=datetime.date.today())
       """

    type = datetime.date
    date_format = "%Y-%m-%d"

    def next_in_enumeration(self, value):
        return value + datetime.timedelta(days=1)

    def normalize(self, value):
        if value is None:
            return None

        if isinstance(value, datetime.datetime):
            value = value.date()
        return value

    def parse_from_str(self, s):
        """
        Parses a date string formatted like ``YYYY-MM-DD``.
        """
        if s.lower() == DateAlias.today:
            return utcnow().date()
        elif s.lower() == DateAlias.yesterday:
            return utcnow().date() - datetime.timedelta(days=1)
        else:
            return datetime.datetime.strptime(s, self.date_format).date()

    def to_str(self, dt):
        """
        Converts the date to a string using the :py:attr:`~_DateValueTypeBase.date_format`.
        """
        if dt is None:
            return str(dt)
        return dt.strftime(self.date_format)


class DateTimeValueType(ValueType):
    """
    DateTimeValueType whose value is a :py:class:`~datetime.datetime` specified to the second.

    A DateSecondValueType is a `ISO 8601 <http://en.wikipedia.org/wiki/ISO_8601>`_ formatted
    date and time specified to the second. For example, ``2013-07-10T190738`` specifies July 10, 2013 at
    19:07:38.
    """

    type = datetime.datetime
    date_format = "%Y-%m-%dT%H%M%S.%f"

    def parse_from_str(self, s):
        """
        Parses a string to a :py:class:`~datetime.datetime`.
        """
        try:
            v = datetime.datetime.strptime(s, self.date_format)
            return v.replace(tzinfo=pytz.UTC)
        except (ValueError):
            return pendulum.parse(s, tz=pytz.UTC)

    def to_str(self, dt):
        """
        Converts the date to a string using the :py:attr:`~_DatetimeValueTypeBase.date_format`.
        """
        if dt is None:
            return str(dt)
        return dt.strftime(self.date_format)

    def normalize(self, dt):
        """
        Clamp dt to every 1th :py:attr:`~DateTimeValueType._timedelta` starting at
        :py:attr:`~_DatetimeValueTypeBase.start`.
        """
        if dt is None:
            return None
        if not isinstance(dt, datetime.datetime):
            dt = datetime.datetime.combine(dt, datetime.time.min)
            dt = dt.replace(tzinfo=pytz.utc)
        return dt



import logging
import pprint

from dbnd._core.tracking.backends.channels.abstract_channel import TrackingChannel


logger = logging.getLogger(__name__)


class ConsoleDebugTrackingChannel(TrackingChannel):
    """Json API client implementation."""

    def __init__(self, print_func=None):
        super(ConsoleDebugTrackingChannel, self).__init__()
        self.print_func = print_func or logger.info
        self._printer = pprint.PrettyPrinter(indent=2, width=140)

    def _handle(self, name, data):
        self.print_func("tracking %s():\n%s" % (name, self._printer.pformat(data)))

    def is_ready(self):
        return True



import logging

from dbnd._core.tracking.backends.channels.abstract_channel import TrackingChannel


logger = logging.getLogger(__name__)


class DisabledTrackingChannel(TrackingChannel):
    def __init__(self):
        super(DisabledTrackingChannel, self).__init__()
        logger.info("Tracking store is disable at core.tracker_api.")

    def is_ready(self):
        return True

    def _handle(self, name, data):
        pass



import os

from contextlib import contextmanager


def environ_enabled(variable_name, default=False):
    # type: (str, Optional[bool]) -> Optional[bool]
    env_value = os.environ.get(variable_name, None)
    if env_value is None:
        return default

    from dbnd._core.utils.basics.helpers import parse_bool

    return parse_bool(env_value)


def environ_int(variable_name, default=None):
    # type: (str, Optional[int]) -> Optional[int]
    env_value = os.environ.get(variable_name, None)
    if env_value is None:
        return default
    try:
        return int(env_value)
    except:
        return default


def set_on(env_key):
    os.environ[env_key] = "True"


def set_env_dir(key, value):
    key = key.upper()
    if key not in os.environ:
        os.environ[key] = os.path.abspath(value)
        return True
    return False


@contextmanager
def env(**environment):
    """
    edit the environment variables only in the context scope
    """
    current = dict(os.environ)
    difference = {key for key in environment if key not in current}

    os.environ.update(environment)
    try:
        yield
    finally:
        os.environ.update(current)
        for key in difference:
            os.environ.pop(key)



import logging


logger = logging.getLogger(__name__)


def target_to_databand(df, target, **kwargs):
    return target.as_pandas.to(df, **kwargs)


try:
    from pandas import DataFrame

    DataFrame.to_target = target_to_databand
except ImportError:
    logger.warning(
        "'pandas' library can not be imported. This will crash at runtime if Pandas functionality is used."
    )

# TODO: Dask
# try:
#     from dask.dataframe import DataFrame
#     DataFrame.to_target = target_to_databand
# except ImportError:
#     logger.warning(
#         "'pandas' library can not be imported. This will crash at runtime if Pandas functionality is used."
#     )



from os import path

import setuptools

from setuptools.config import read_configuration


BASE_PATH = path.dirname(__file__)
CFG_PATH = path.join(BASE_PATH, "setup.cfg")

config = read_configuration(CFG_PATH)

# A list of vendored packages
dbnd_vendors_list = [
    "cachetools",
    "hjson",
    "cloudpickle",
    "pendulum==1.4.4",
    "tabulate",
    "marshmallow==2.18.0",
    "croniter>=0.3.30,<0.4",
    "protobuf==3.13.0",
    "psutil>=4.2.0,<5.7.0",  # extracted use to vendorized_psutil.py
]


setuptools.setup(
    name="dbnd",
    package_dir={"": "src"},
    install_requires=[
        "tzlocal>=1.0.0,<2.0.0",
        "six",
        "more_itertools",
        "attrs",
        "pyyaml",  # yaml support in targets
        "pytz",  # python time zone for pendulum library
        "pytzdata",  # python time zone for pendulum library
        "requests>=2.18.0",  # API TRACKING
        "configparser<3.6.0,>=3.5.0",  # same versions as Airflow -- Moved library to _vendor
        "jinja2>=2.10.1, <2.11.0",  # same versions as Airflow
        "gitpython",
        'pygments<=2.5.2 ; python_version < "3.0"',
        'pygments>=2.6.1 ; python_version >= "3.0"',
        # backward compatible python
        'typing;python_version<"3.7"',  # Standalone pkg is incompatible with 3.7 and not required
        'pathlib2; python_version < "3.0"',  # pathlib support in python 2
        "pathlib2;python_version<='2.7'",
        "enum34;python_version<='2.7'",
        'contextlib2; python_version < "3"',
        "python-dateutil",
    ],
    extras_require={
        ':sys_platform=="win32"': ["colorama"],
        "tests": [
            "qtconsole==4.7.7",  # 5.0 is not py 27 compatible
            "pandas==0.24.2",
            "numpy",
            "coverage",
            "pytest==4.5.0",  # 4.6.0 requires pluggy 0.12
            "pytest-cov==2.9.0",
            "pluggy==0.11.0",  # 0.12 has import_metadata, fails on py2
            "zope.interface",
            "mock",
            "pandas<2.0.0,>=0.17.1",  # airflow supports only this version
            "urllib3==1.23",  # otherwise we have 1.24 - conflict with 'requests'
            "tox==3.12.1",
            "matplotlib==2.2.5;python_version<'3.5'",
            "matplotlib==3.3.0;python_version>='3.5'",
            "tables==3.5.1",
            "feather-format",
            "pyarrow",
            "nbconvert",
            "nbformat",
            "jupyter",
            "traitlets>=4.2,<5.0.0",  # required by jupyter, fix py37 compatibility
            "IPython>=4.0.0, <7.0",
            "jupyter_contrib_nbextensions",
            "idna<=2.7",  # conflict with requests (require 2.8 <)
            # conflict with pandas version on new openpyxl: got invalid input value of type <class 'xml.etree.ElementTree.Element'>, expected string or Element
            "openpyxl==2.6.4",
            "sklearn",
            "WTForms<2.3.0",  # fixing ImportError: cannot import name HTMLString at 2.3.0
            "wheel",  # for fat_wheel tests
        ],
    },
    entry_points={"console_scripts": ["dbnd = dbnd:dbnd_main"]},
)



import contextlib
import logging
import os
import sys
import threading
import typing

from datetime import datetime
from typing import Any, ContextManager, Iterator, List, Optional, Union
from uuid import UUID

import six

from dbnd._core.configuration.environ_config import (
    DBND_PARENT_TASK_RUN_ATTEMPT_UID,
    DBND_PARENT_TASK_RUN_UID,
    DBND_RESUBMIT_RUN,
    DBND_ROOT_RUN_TRACKER_URL,
    DBND_ROOT_RUN_UID,
    DBND_RUN_UID,
    ENV_DBND__USER_PRE_INIT,
)
from dbnd._core.constants import (
    DescribeFormat,
    RunState,
    SystemTaskName,
    TaskExecutorType,
    TaskRunState,
    UpdateSource,
)
from dbnd._core.current import (
    current_task_run,
    try_get_current_task,
    try_get_databand_run,
)
from dbnd._core.errors import DatabandRuntimeError
from dbnd._core.errors.base import (
    DatabandFailFastError,
    DatabandRunError,
    DatabandSigTermError,
)
from dbnd._core.parameter.parameter_builder import output, parameter
from dbnd._core.plugin.dbnd_plugins import is_airflow_enabled
from dbnd._core.run.describe_run import DescribeRun, print_tasks_tree
from dbnd._core.run.run_tracker import RunTracker
from dbnd._core.run.target_identity_source_map import TargetIdentitySourceMap
from dbnd._core.run.task_runs_builder import TaskRunsBuilder
from dbnd._core.settings import DatabandSettings, EngineConfig, RunConfig
from dbnd._core.task import Task
from dbnd._core.task_build.task_context import current_task, has_current_task
from dbnd._core.task_build.task_registry import (
    build_task_from_config,
    get_task_registry,
)
from dbnd._core.task_executor.factory import (
    calculate_task_executor_type,
    get_task_executor,
)
from dbnd._core.task_executor.heartbeat_sender import start_heartbeat_sender
from dbnd._core.task_run.task_run import TaskRun
from dbnd._core.tracking.schemas.tracking_info_run import RootRunInfo, ScheduledRunInfo
from dbnd._core.utils import console_utils
from dbnd._core.utils.basics.load_python_module import load_python_callable
from dbnd._core.utils.basics.nested_context import nested
from dbnd._core.utils.basics.singleton_context import SingletonContext
from dbnd._core.utils.date_utils import unique_execution_date
from dbnd._core.utils.timezone import utcnow
from dbnd._core.utils.traversing import flatten
from dbnd._core.utils.uid_utils import get_uuid
from dbnd._vendor.cloudpickle import cloudpickle
from dbnd._vendor.namesgenerator import get_name_for_uid
from dbnd.api.runs import kill_run
from targets import FileTarget, Target
from targets.caching import TARGET_CACHE


logger = logging.getLogger(__name__)

AD_HOC_DAG_PREFIX = "DBND_RUN."


def _get_dbnd_run_relative_cmd():
    argv = list(sys.argv)
    while argv:
        current = argv.pop(0)
        if current == "run":
            return argv
    raise DatabandRunError(
        "Can't calculate run command from '%s'",
        help_msg="Check that it has a format of '..executable.. run ...'",
    )


# naive implementation of stop event
# we can't save it on Context (non pickable in some cases like running in multithread python)
# if somebody is killing run it's global for the whole process
_is_killed = threading.Event()


class DatabandRun(SingletonContext):
    def __init__(
        self,
        context,
        task_or_task_name,
        run_uid=None,
        scheduled_run_info=None,
        send_heartbeat=True,
        existing_run=None,
        job_name=None,
        source=UpdateSource.dbnd,
        af_context=None,
    ):
        # type:(DatabandContext, Union[Task, str] , Optional[UUID], Optional[ScheduledRunInfo], Optional[bool], Optional[UpdateSource]) -> None
        self.context = context
        s = self.context.settings  # type: DatabandSettings

        if isinstance(task_or_task_name, six.string_types):
            self.root_task_name = task_or_task_name
            self.root_task = None
        elif isinstance(task_or_task_name, Task):
            self.root_task_name = task_or_task_name.task_name
            self.root_task = task_or_task_name
        else:
            raise

        self.job_name = job_name or self.root_task_name

        self.description = s.run.description
        self.is_archived = s.run.is_archived
        self.source = source
        # this was added to allow the scheduler to create the run which will be continued by the actually run command instead of having 2 separate runs
        if not run_uid and DBND_RUN_UID in os.environ:
            # we pop so if this run spawnes subprocesses with their own runs they will be associated using the sub-runs mechanism instead
            # of being fused into this run directly
            run_uid = os.environ.pop(DBND_RUN_UID)
        if run_uid:
            self.run_uid = run_uid
            self.existing_run = True
        else:
            self.run_uid = get_uuid()
            self.existing_run = False

        if existing_run is not None:
            self.existing_run = existing_run

        self.name = s.run.name or get_name_for_uid(self.run_uid)
        # this is so the scheduler can create a run with partial information and then have the subprocess running the actual cmd fill in the details
        self.resubmit_run = (
            DBND_RESUBMIT_RUN in os.environ
            and os.environ.pop(DBND_RESUBMIT_RUN) == "true"
        )

        # AIRFLOW, move into executor
        # dag_id , execution_date and run_id is used by airflow
        self.dag_id = AD_HOC_DAG_PREFIX + self.root_task_name
        self.execution_date = unique_execution_date()
        run_id = s.run.id
        if not run_id:
            # we need this name, otherwise Airflow will try to manage our local jobs at scheduler
            # ..zombies cleanup and so on
            run_id = "backfill_{0}_{1}".format(
                self.name, self.execution_date.isoformat()
            )
        self.run_id = run_id

        self._template_vars = self._build_template_vars()

        self.is_tracked = True

        self.runtime_errors = []
        self._run_state = None
        self.task_runs = []  # type: List[TaskRun]
        self.task_runs_by_id = {}
        self.task_runs_by_af_id = {}

        self.target_origin = TargetIdentitySourceMap()
        self.describe = DescribeRun(self)
        self.tracker = RunTracker(self, tracking_store=self.context.tracking_store)

        # ALL RUN CONTEXT SPECIFIC thing
        self.root_run_info = RootRunInfo.from_env(current_run=self)
        self.scheduled_run_info = scheduled_run_info or ScheduledRunInfo.from_env(
            self.run_uid
        )

        # now we can add driver task
        self.driver_task_run = None  # type: Optional[TaskRun]
        self.root_task_run = None  # type: Optional[TaskRun]

        self.run_folder_prefix = os.path.join(
            "log",
            self.execution_date.strftime("%Y-%m-%d"),
            "%s_%s_%s"
            % (
                self.execution_date.strftime("%Y-%m-%dT%H%M%S.%f"),
                self.root_task_name,
                self.name,
            ),
        )

        self.run_config = self.context.settings.run  # type: RunConfig
        self.env = env = self.context.env

        self.local_engine = self._get_engine_config(env.local_engine)
        self.remote_engine = self._get_engine_config(
            env.remote_engine or env.local_engine
        )

        self.submit_driver = (
            self.run_config.submit_driver
            if self.run_config.submit_driver is not None
            else env.submit_driver
        )
        self.submit_tasks = (
            self.run_config.submit_tasks
            if self.run_config.submit_tasks is not None
            else env.submit_tasks
        )
        self.task_executor_type, self.parallel = calculate_task_executor_type(
            self.submit_tasks, self.remote_engine, self.context.settings
        )

        self.sends_heartbeat = send_heartbeat
        self.dynamic_af_tasks_count = dict()
        self.af_context = af_context
        self.start_time = None
        self.finished_time = None

    def _get_engine_config(self, name):
        # type: ( Union[str, EngineConfig]) -> EngineConfig
        return build_task_from_config(name, EngineConfig)

    @property
    def run_url(self):
        return self.tracker.run_url

    @property
    def task(self):
        return self.root_task

    @property
    def driver_task(self):
        # type: ()->_DbndDriverTask
        return self.driver_task_run.task

    @property
    def driver_dump(self):
        return self.driver_task_run.task.driver_dump

    def _build_template_vars(self):
        # template vars
        ds = self.execution_date.strftime("%Y-%m-%d")
        ts = self.execution_date.isoformat()
        ds_nodash = ds.replace("-", "")
        ts_nodash = ts.replace("-", "").replace(":", "")
        ts_safe = ts.replace(":", "")

        return {
            "run": self,
            "run_ds": ds,
            "run_ts": ts,
            "run_ds_nodash": ds_nodash,
            "run_ts_nodash": ts_nodash,
            "run_ts_safe": ts_safe,
        }

    # TODO: split to get_by_id/by_af_id
    def get_task_run(self, task_id):
        # type: (str) -> TaskRun
        return self.get_task_run_by_id(task_id) or self.get_task_run_by_af_id(task_id)

    def get_task_run_by_id(self, task_id):
        # type: (str) -> TaskRun
        return self.task_runs_by_id.get(task_id)

    def get_task_run_by_af_id(self, task_id):
        # type: (str) -> TaskRun
        return self.task_runs_by_af_id.get(task_id)

    def get_af_task_ids(self, task_ids):
        return [self.get_task_run(task_id).task_af_id for task_id in task_ids]

    def get_task(self, task_id):
        # type: (str) -> Task
        return self.get_task_run(task_id).task

    @property
    def describe_dag(self):
        return self.root_task.ctrl.describe_dag

    def set_run_state(self, state):
        self._run_state = state
        self.tracker.set_run_state(state)

    def run_dynamic_task(self, task, task_engine=None):
        if task_engine is None:
            task_engine = self.current_engine_config
        task_run = self.create_dynamic_task_run(task, task_engine)
        task_run.runner.execute()
        return task_run

    def _build_driver_task(self):
        if self.submit_driver and not self.existing_run:
            logger.info("Submitting job to remote execution")
            task_name = SystemTaskName.driver_submit
            is_submitter = True
            is_driver = False
            host_engine = self.local_engine.clone(require_submit=False)
            target_engine = self.local_engine.clone(require_submit=False)
            task_executor_type = TaskExecutorType.local
        else:
            task_name = SystemTaskName.driver
            # even if it's existing run, may be we are running from Airflow
            # so the run is actually "submitted", ( the root airflow job has no info..,
            # we want to capture "real" info of the run
            is_submitter = not self.existing_run or self.resubmit_run
            is_driver = True
            task_executor_type = self.task_executor_type

            if self.submit_driver:
                # submit drive is true, but we are in existing run:
                # we are after the jump from submit to driver execution (to remote engine)
                host_engine = self.remote_engine.clone(require_submit=False)
            else:
                host_engine = self.local_engine.clone(
                    require_submit=False
                )  # we are running at this engine already

            target_engine = self.remote_engine
            if not self.submit_tasks or task_executor_type == "airflow_kubernetes":
                target_engine = target_engine.clone(require_submit=False)

        dbnd_local_root = host_engine.dbnd_local_root or self.env.dbnd_local_root
        run_folder_prefix = self.run_folder_prefix

        local_driver_root = dbnd_local_root.folder(run_folder_prefix)
        local_driver_log = local_driver_root.partition("%s.log" % task_name)

        remote_driver_root = self.env.dbnd_root.folder(run_folder_prefix)
        driver_dump = remote_driver_root.file("%s.pickle" % task_name)

        driver_task = _DbndDriverTask(
            task_name=task_name,
            task_version=self.run_uid,
            execution_date=self.execution_date,
            is_submitter=is_submitter,
            is_driver=is_driver,
            host_engine=host_engine,
            target_engine=target_engine,
            task_executor_type=task_executor_type,
            local_driver_root=local_driver_root,
            local_driver_log=local_driver_log,
            remote_driver_root=remote_driver_root,
            driver_dump=driver_dump,
            sends_heartbeat=self.sends_heartbeat,
        )

        tr = TaskRun(task=driver_task, run=self, task_engine=driver_task.host_engine)
        self._add_task_run(tr)
        return tr

    def _on_enter(self):
        if self.driver_task_run is None:
            # we are in submit/driver
            self.driver_task_run = self._build_driver_task()
            self.current_engine_config = self.driver_task_run.task.host_engine
            self.tracker.init_run()
        else:
            # we are in task run ( after the jump)
            self.current_engine_config = self.driver_task_run.task.target_engine.clone(
                require_submit=False
            )

    def _dbnd_run_error(self, ex):
        if (
            # what scenario is this aiflow filtering supposed to help with?
            # I had airflow put a default airflow.cfg in .dbnd causing validation error in k8sExecutor which was invisible in the console (only in task log)
            (
                "airflow" not in ex.__class__.__name__.lower()
                or ex.__class__.__name__ == "AirflowConfigException"
            )
            and "Failed tasks are:" not in str(ex)
            and not isinstance(ex, DatabandRunError)
            and not isinstance(ex, KeyboardInterrupt)
            and not isinstance(ex, DatabandSigTermError)
        ):
            logger.exception(ex)

        if (
            isinstance(ex, KeyboardInterrupt)
            or isinstance(ex, DatabandSigTermError)
            or self.is_killed()
        ):
            run_state = RunState.CANCELLED
            unfinished_task_state = TaskRunState.UPSTREAM_FAILED
        elif isinstance(ex, DatabandFailFastError):
            run_state = RunState.FAILED
            unfinished_task_state = TaskRunState.UPSTREAM_FAILED
        else:
            run_state = RunState.FAILED
            unfinished_task_state = TaskRunState.FAILED

        self.set_run_state(run_state)
        self.tracker.tracking_store.set_unfinished_tasks_state(
            run_uid=self.run_uid, state=unfinished_task_state
        )

        err_banner_msg = self.describe.get_error_banner()
        logger.error(
            u"\n\n{sep}\n{banner}\n{sep}".format(
                sep=console_utils.ERROR_SEPARATOR, banner=err_banner_msg
            )
        )
        return DatabandRunError(
            "Run has failed: %s" % ex, run=self, nested_exceptions=ex
        )

    def run_driver(self):
        """
        Runs the main driver!
        """
        # with captures_log_into_file_as_task_file(log_file=self.local_driver_log.path):
        try:
            self.start_time = utcnow()
            logger.info("Running driver... Driver PID: %s", os.getpid())
            self.driver_task_run.runner.execute()
            self.finished_time = utcnow()
        except DatabandRunError as ex:
            self._dbnd_run_error(ex)
            raise
        except (Exception, KeyboardInterrupt, SystemExit) as ex:
            raise self._dbnd_run_error(ex)
        finally:
            try:
                self.driver_task.host_engine.cleanup_after_run()
            except Exception:
                logger.exception("Failed to shutdown the current run, continuing")

        return self

    @property
    def duration(self):
        if self.finished_time and self.start_time:
            return self.finished_time - self.start_time
        return None

    def _get_task_by_id(self, task_id):
        task = self.context.task_instance_cache.get_task_by_id(task_id)
        if task is None:
            raise DatabandRuntimeError(
                "Failed to find task %s in current context" % task_id
            )

        return task

    def save_run(self, target_file=None):
        """
        dumps current run and context to file
        """
        t = target_file or self.driver_dump
        logger.info("Saving current run into %s", t)
        with t.open("wb") as fp:
            cloudpickle.dump(obj=self, file=fp)

    @contextlib.contextmanager
    def run_context(self):
        # type: (DatabandRun) -> Iterator[DatabandRun]

        from dbnd._core.context.databand_context import DatabandContext  # noqa: F811

        with DatabandContext.context(_context=self.context):
            with DatabandRun.context(_context=self) as dr:
                yield dr

    @classmethod
    def load_run(cls, dump_file, disable_tracking_api):
        # type: (FileTarget, bool) -> DatabandRun
        logger.info("Loading dbnd run from %s", dump_file)
        with dump_file.open("rb") as fp:
            databand_run = cloudpickle.load(file=fp)
            if disable_tracking_api:
                databand_run.context.tracking_store.disable_tracking_api()
                logger.info("Tracking has been disabled")
        try:
            if databand_run.context.settings.core.pickle_handler:
                pickle_handler = load_python_callable(
                    databand_run.context.settings.core.pickle_handler
                )
                pickle_handler(databand_run)
        except Exception as e:
            logger.warning(
                "error while trying to handle pickle with custom handler:", e
            )
        return databand_run

    def get_template_vars(self):
        return self._template_vars

    def create_dynamic_task_run(self, task, task_engine, task_af_id=None):
        if task_af_id is None:
            task_name = task.friendly_task_name
            if task_name in self.dynamic_af_tasks_count:
                self.dynamic_af_tasks_count[task_name] += 1
                task_af_id = "{}_{}".format(
                    task_name, self.dynamic_af_tasks_count[task_name]
                )
            else:
                self.dynamic_af_tasks_count[task_name] = 1
                task_af_id = task_name

        tr = TaskRun(
            task=task,
            run=self,
            is_dynamic=True,
            task_engine=task_engine,
            task_af_id=task_af_id,
        )
        self.add_task_runs([tr])
        return tr

    def add_task_runs(self, task_runs):
        # type: (List[TaskRun]) -> None
        for tr in task_runs:
            self._add_task_run(tr)

        self.tracker.add_task_runs(task_runs)

    def _add_task_run(self, task_run):
        self.task_runs.append(task_run)
        self.task_runs_by_id[task_run.task.task_id] = task_run
        self.task_runs_by_af_id[task_run.task_af_id] = task_run

        task_run.task.ctrl.last_task_run = task_run

    def cleanup_after_task_run(self, task):
        # type: (Task) -> None
        rels = task.ctrl.relations
        # potentially, all inputs/outputs targets for current task could be removed
        targets_to_clean = set(flatten([rels.task_inputs, rels.task_outputs]))

        targets_in_use = set()
        # any target which appears in inputs of all not finished tasks shouldn't be removed
        for tr in self.task_runs:
            if tr.task_run_state in TaskRunState.final_states():
                continue
            # remove all still needed inputs from targets_to_clean list
            for target in flatten(tr.task.ctrl.relations.task_inputs):
                targets_in_use.add(target)

        TARGET_CACHE.clear_for_targets(targets_to_clean - targets_in_use)

    def get_context_spawn_env(self):
        env = {}
        if has_current_task():
            current = current_task()
        else:
            current = self.root_task

        if current:
            tr = self.get_task_run_by_id(current.task_id)
            if tr:
                env[DBND_PARENT_TASK_RUN_UID] = str(tr.task_run_uid)
                env[DBND_PARENT_TASK_RUN_ATTEMPT_UID] = str(tr.task_run_attempt_uid)

        env[DBND_ROOT_RUN_UID] = str(self.root_run_info.root_run_uid)
        env[DBND_ROOT_RUN_TRACKER_URL] = self.root_run_info.root_run_url

        if self.context.settings.core.user_code_on_fork:
            env[ENV_DBND__USER_PRE_INIT] = self.context.settings.core.user_code_on_fork
        return env

    def _init_without_run(self):
        self.driver_task_run.task.build_root_task_runs(self)

    def is_killed(self):
        return _is_killed.is_set()

    def _internal_kill(self):
        """
        called by TaskRun handler, so we know that run is "canceled"
        otherwise we will get regular exception
        """
        _is_killed.set()

    def kill(self):
        """
        called from user space, kills the current task only
        :return:
        """
        # this is very naive stop implementation
        # in case of simple executor, we'll run task.on_kill code
        _is_killed.set()
        try:
            current_task = None
            from dbnd._core.task_build.task_context import TaskContext, TaskContextPhase

            tc = TaskContext.try_instance()
            if tc.phase == TaskContextPhase.RUN:
                current_list = list(tc.stack)
                if current_list:
                    current_task = current_list.pop()
        except Exception as ex:
            logger.error("Failed to find current task: %s" % ex)
            return

        if not current_task:
            logger.info("No current task.. Killing nothing..")
            return

        try:
            current_task.on_kill()
        except Exception as ex:
            logger.error("Failed to kill current task %s: %s" % (current_task, ex))
            return

    def kill_run(self):
        _is_killed.set()
        try:
            return kill_run(str(self.run_uid), ctx=self.context)
        except Exception as e:
            raise DatabandFailFastError(
                "Could not send request to kill databand run!", e
            )

    def get_current_dbnd_local_root(self):
        # we should return here the proper engine config, based in which context we run right now
        # it could be submit, driver or task engine
        return self.env.dbnd_local_root

    @property
    def is_orchestration(self):
        return self.source == UpdateSource.dbnd


class _DbndDriverTask(Task):
    _conf__no_child_params = True
    task_is_system = True
    task_in_memory_outputs = True

    is_driver = parameter[bool]
    is_submitter = parameter[bool]
    execution_date = parameter[datetime]
    sends_heartbeat = parameter[bool]

    host_engine = parameter[EngineConfig]
    target_engine = parameter[EngineConfig]

    task_executor_type = parameter[str]

    # all paths, we make them system, we don't want to check if they are exists
    local_driver_root = output(system=True)[Target]
    local_driver_log = output(system=True)[Target]

    remote_driver_root = output(system=True)[Target]
    driver_dump = output(system=True)[Target]

    def _build_submit_task(self, run):
        if run.root_task:
            raise DatabandRuntimeError(
                "Can't send to remote execution task created via code, only command line is supported"
            )

        # dont' describe in local run, do it in remote run
        settings = self.settings
        settings.system.describe = False

        cmd_line_args = (
            ["run"] + _get_dbnd_run_relative_cmd() + ["--run-driver", str(run.run_uid)]
        )

        args = run.remote_engine.dbnd_executable + cmd_line_args

        root_task = run.remote_engine.submit_to_engine_task(
            env=run.env,
            args=args,
            task_name="dbnd_driver_run",
            interactive=settings.run.interactive,
        )
        root_task._conf_confirm_on_kill_msg = (
            "Ctrl-C Do you want to kill your submitted pipeline?"
            "If selection is 'no', this process will detach from the run."
        )
        return root_task

    def _build_root_task(self, run):
        # type: (DatabandRun) -> Task
        if self.is_submitter and not self.is_driver:
            return self._build_submit_task(run)
        else:
            if run.root_task:
                # user has created DatabandRun with existing task
                self.task_meta.add_child(run.root_task.task_id)
                return run.root_task

            logger.info("Building main task '%s'", run.root_task_name)
            root_task = get_task_registry().build_dbnd_task(run.root_task_name)
            logger.info(
                "Task %s has been created (%s children)",
                root_task.task_id,
                len(root_task.ctrl.task_dag.subdag_tasks()),
            )
            return root_task

    def is_save_run(self, run, task_runs):

        core_settings = run.context.settings.core
        if core_settings.always_save_pipeline:
            return True
        if core_settings.disable_save_pipeline:
            return False

        if any(tr.task._conf__require_run_dump_file for tr in task_runs):
            return True

        if self.target_engine.require_submit:
            return True

        if self.task_executor_type == TaskExecutorType.local:
            return False

        if is_airflow_enabled():
            from dbnd_airflow.executors import AirflowTaskExecutorType

            return self.task_executor_type not in [
                AirflowTaskExecutorType.airflow_inprocess,
                TaskExecutorType.local,
            ]
        return True

    def build_task_from_cmd_line(self, task_name):
        return

    def build_root_task_runs(self, run):
        """
        called by .run and inline
        :return:
        """
        run.root_task = self._build_root_task(run)

        # for validation only
        run.root_task.task_dag.topological_sort()

        task_runs = TaskRunsBuilder().build_task_runs(
            run, run.root_task, remote_engine=self.target_engine
        )
        # we need it before to mark root task
        run.add_task_runs(task_runs)
        # for faster access
        run.root_task_run = run.get_task_run(run.root_task.task_id)
        return task_runs

    def run(self):
        driver_task_run = current_task_run()
        run = driver_task_run.run  # type: DatabandRun
        if self.is_submitter:
            run.set_run_state(RunState.RUNNING)

        ctx = run.context
        ctx.settings.git.validate_git_policy()

        # let prepare for remote execution
        run.remote_engine.prepare_for_run(run)

        task_runs = self.build_root_task_runs(run)

        hearbeat = None

        # right now we run describe in local controller only, but we should do that for more
        if self.is_driver:
            if run.context.settings.system.describe:
                run.describe_dag.describe_dag()
                logger.info(run.describe.run_banner("Described!", color="blue"))
                return

            root_task_run = run.root_task_run
            run.root_task.ctrl.banner(
                "Main task '%s' has been created!" % root_task_run.task_af_id,
                color="cyan",
                task_run=root_task_run,
            )

            print_tasks_tree(root_task_run.task, task_runs)

            if self.is_save_run(run, task_runs):
                run.save_run()

            if self.sends_heartbeat:
                hearbeat = start_heartbeat_sender(driver_task_run)

        task_runs_to_run = [tr for tr in task_runs if not tr.is_skipped]

        # create executor without driver task!
        task_executor = get_task_executor(
            run,
            task_executor_type=self.task_executor_type,
            host_engine=self.host_engine,
            target_engine=run.root_task_run.task_engine,
            task_runs=task_runs_to_run,
        )

        with nested(hearbeat):
            task_executor.do_run()

        if self.is_driver:
            # This is great success!
            run.set_run_state(RunState.SUCCESS)
            logger.info(run.describe.run_banner_for_finished())
            return run
        else:
            logger.info(run.describe.run_banner_for_submitted())


def new_databand_run(context, task_or_task_name, run_uid=None, **kwargs):
    # type: (DatabandContext, Union[Task, str], UUID, **Any)-> ContextManager[DatabandRun]

    kwargs["allow_override"] = kwargs.pop("allow_override", True)
    return DatabandRun.new_context(
        context=context, task_or_task_name=task_or_task_name, run_uid=run_uid, **kwargs
    )



from functools import reduce, wraps
from operator import xor

from dbnd._vendor import click


class NotRequiredIf(click.Option):
    def __init__(self, *args, **kwargs):
        self.not_required_if = kwargs.pop("not_required_if")
        assert self.not_required_if, "'not_required_if' parameter required"
        kwargs["help"] = (
            kwargs.get("help", "")
            + " NOTE: This argument is mutually exclusive with %s"
            % self.not_required_if
        ).strip()
        super(NotRequiredIf, self).__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        we_are_present = self.name in opts
        other_present = self.not_required_if in opts

        if other_present:
            if we_are_present:
                raise click.UsageError(
                    "Illegal usage: `%s` is mutually exclusive with `%s`"
                    % (self.name, self.not_required_if)
                )
            else:
                self.prompt = None

        return super(NotRequiredIf, self).handle_parse_result(ctx, opts, args)


def only_one(*args):
    return reduce(xor, map(bool, args))


def requierd_mutually_exclude_options(*options):
    """
    Creates a decorator for click commands that make sure that
    only one, and at least one, of `options` is set while calling the click command.
    It also set a relevant help message for the mutually excluded options.
    """

    def decorator(func):
        options_obj = [
            option for option in func.__click_params__ if option.name in options
        ]
        options_names = {option.name for option in options_obj}
        for option in options_obj:
            current_help = "" if option.help is None else option.help + " "
            others = list(options_names - {option.name})
            added_msg = "**mutually excluded with {others}".format(others=others)
            option.help = current_help + added_msg

        @wraps(func)
        def inner(**kwargs):
            relevant_options = [
                value for name, value in kwargs.items() if name in options
            ]
            if not only_one(*relevant_options):
                raise click.UsageError(
                    "Illegal usage: `at least one, and only one of the options [{options}] is required` ".format(
                        options=",".join("--" + op_name for op_name in options)
                    )
                )
            func(**kwargs)

        return inner

    return decorator


def options_dependency(root, *dependents):
    """
    Creates a decorator for click commands that make sure that only when `root`
    option is set all the others `dependents` could be used.
    It also set a relevant help message for the dependents options.
    """

    def decorator(func):
        options_obj = [
            option for option in func.__click_params__ if option.name in dependents
        ]
        for option in options_obj:
            current_help = "" if option.help is None else option.help + " "
            added_msg = "**depend on {root} option".format(root=root)
            option.help = current_help + added_msg

        @wraps(func)
        def inner(**kwargs):
            relevant_options = [
                value for name, value in kwargs.items() if name in dependents
            ]

            if any(relevant_options) and kwargs.get(root) is None:
                raise click.UsageError(
                    "Illegal usage: `in oprder to use any of this options {options}, "
                    "the {root} option must be set".format(
                        root=root, options=dependents
                    )
                )

            func(**kwargs)

        return inner

    return decorator



import logging
import re

import six

from six import StringIO

from dbnd._core.configuration.environ_config import ENV_DBND__NO_TABLES
from dbnd._core.utils import json_utils
from dbnd._core.utils.basics.environ_utils import environ_enabled
from dbnd._core.utils.platform import windows_compatible_mode
from dbnd._core.utils.string_utils import safe_short_string
from dbnd._core.utils.terminal import get_terminal_size
from dbnd._core.utils.traversing import traverse, traverse_to_str
from dbnd._vendor.tabulate import tabulate
from dbnd._vendor.termcolor import colored


logger = logging.getLogger(__name__)

_TAB = "\n\t      "


def safe_string(value, max_value_len=1000):
    try:
        if not isinstance(value, six.string_types):
            value = str(value)

        return safe_short_string(value=value, max_value_len=max_value_len)
    except Exception as ex:
        return "Failed to convert value to string:%s" % ex


class TextBanner(StringIO):
    def __init__(self, msg, color="white"):
        StringIO.__init__(self)

        self.color = color
        self._banner_separator = colored("==================== \n", color)
        self._banner_msg = colored("= %s\n" % msg, color)
        self._section_separator = "-------\n"
        # p = pprint.PrettyPrinter(indent=2, width=140)

        self.write("\n")
        self.write(self._banner_separator)
        self.write(self._banner_msg)

    @staticmethod
    def f_io(structure):

        structure_str = traverse_to_str(structure)
        structure_str = traverse(
            structure_str,
            lambda x: x
            if not x or len(x) <= 600
            else ("%s... (%s files)" % (x[:400], len(x.split(",")))),
        )
        dumped = json_utils.dumps(structure_str, indent=2)
        return dumped

    def f_struct(self, structure):
        # return p.pformat(_dump_struct(structure))
        structure_str = traverse_to_str(structure)
        dumped = json_utils.dumps(structure_str, indent=2)
        return dumped

    def f_simple_dict(self, params, skip_if_empty=False):
        return "  ".join(
            "%s=%s" % (key, value)
            for key, value in params
            if not skip_if_empty or value is not None
        )

    def column_properties(self, name, params, skip_if_empty=False, **kwargs):
        return self.column(
            name,
            self.f_simple_dict(params, skip_if_empty=skip_if_empty),
            skip_if_empty=skip_if_empty,
            **kwargs
        )

    def column(self, name, value, newline=True, raw_name=False, skip_if_empty=False):
        # if separate:
        #     s.write("  \n")
        if skip_if_empty and value is None:
            return self

        if isinstance(value, six.string_types):
            if six.PY2:
                value = value.encode("utf-8", errors="ignore")
            else:
                value = str(value)
        else:
            value = str(value)

        if skip_if_empty and not value:
            return self

        if "\n" in value:
            value = _TAB + _TAB.join(value.split("\n"))

        # remove last \n,  use new line or new section if you nee that
        if value[-1:] == "\n":
            value = value[:-1]

        self.write(
            " {name:<18} : {value}".format(
                name=name if raw_name else colored(name, attrs=["bold"]), value=value
            )
        )

        if newline:
            self.write("\n")
        return self

    def new_section(self):
        self.write(self._section_separator)

    def new_line(self):
        self.write("\n")

    def get_banner_str(self):
        self.write(colored("=\n", self.color))
        self.write(self._banner_separator)
        return self.getvalue()


def safe_tabulate(tabular_data, headers, **kwargs):
    terminal_columns, _ = get_terminal_size()
    # fancy_grid format has utf-8 characters (in corners of table)
    # cp1252 fails to encode that
    fancy_grid = not windows_compatible_mode and not environ_enabled(
        ENV_DBND__NO_TABLES
    )
    tablefmt = "fancy_grid" if fancy_grid else "grid"
    table = tabulate(tabular_data, headers=headers, tablefmt=tablefmt, **kwargs)
    if table and max(map(len, table.split())) >= terminal_columns:
        table = tabulate(tabular_data, headers=headers, tablefmt="plain", **kwargs)
    return table



import datetime

from typing import Dict, List, Optional

import attr

from dbnd._core.tracking.schemas.base import _ApiCallSchema
from dbnd._vendor.marshmallow import fields, post_load


class AirflowServerInfoSchema(_ApiCallSchema):
    base_url = fields.String()
    external_url = fields.String(allow_none=True)
    airflow_instance_uid = fields.String(allow_none=True)
    airflow_version = fields.String(allow_none=True)
    airflow_export_version = fields.String(allow_none=True)
    airflow_monitor_version = fields.String(allow_none=True)
    dags_path = fields.String(allow_none=True)
    logs_path = fields.String(allow_none=True)
    last_sync_time = fields.DateTime(allow_none=True)
    monitor_status = fields.String(allow_none=True)
    monitor_error_message = fields.String(allow_none=True)
    monitor_start_time = fields.DateTime(allow_none=True)
    synced_from = fields.DateTime(allow_none=True)
    synced_to = fields.DateTime(allow_none=True)
    api_mode = fields.String(allow_none=True)
    sync_interval = fields.Integer(allow_none=True)
    is_sync_enabled = fields.Boolean(allow_none=True)
    fetcher = fields.String(allow_none=True)
    composer_client_id = fields.String(allow_none=True)
    active_dags = fields.Dict(allow_none=True)
    name = fields.String(allow_none=True)
    env = fields.String(allow_none=True)

    @post_load
    def make_object(self, data, **kwargs):
        return AirflowServerInfo(**data)


airflow_server_info_schema = AirflowServerInfoSchema()


@attr.s
class AirflowServerInfo(object):
    base_url = attr.ib()  # type: str
    external_url = attr.ib(default=None)  # type: Optional[str]
    airflow_instance_uid = attr.ib(default=None)  # type: str
    airflow_version = attr.ib(default=None)  # type: Optional[str]
    airflow_export_version = attr.ib(default=None)  # type: Optional[str]
    airflow_monitor_version = attr.ib(default=None)  # type: Optional[str]
    dags_path = attr.ib(default=None)  # type: Optional[str]
    logs_path = attr.ib(default=None)  # type: Optional[str]
    last_sync_time = attr.ib(default=None)  # type: Optional[datetime.datetime]
    monitor_status = attr.ib(default=None)  # type: Optional[str]
    monitor_error_message = attr.ib(default=None)  # type: Optional[str]
    monitor_start_time = attr.ib(default=None)  # type: Optional[datetime.datetime]
    synced_from = attr.ib(default=None)  # type: Optional[datetime.datetime]
    synced_to = attr.ib(default=None)  # type: Optional[datetime.datetime]
    api_mode = attr.ib(default="rbac")  # type: Optional[str]
    sync_interval = attr.ib(default=None)  # type: Optional[int]
    is_sync_enabled = attr.ib(default=None)  # type: Optional[bool]
    fetcher = attr.ib(default=None)  # type: Optional[str]
    composer_client_id = attr.ib(default=None)  # type: Optional[str]
    active_dags = attr.ib(default=None)  # type: Dict[str, List[str]]
    name = attr.ib(default=None)  # type: Optional[str]
    env = attr.ib(default=None)  # type: Optional[str]



import datetime
import typing

from uuid import UUID

import attr

from dbnd._core.constants import (
    DbndTargetOperationStatus,
    DbndTargetOperationType,
    RunState,
    TaskRunState,
    UpdateSource,
)
from dbnd._core.inplace_run.airflow_dag_inplace_tracking import AirflowTaskContext
from dbnd._core.tracking.schemas.base import ApiObjectSchema, _ApiCallSchema
from dbnd._core.tracking.schemas.metrics import Metric
from dbnd._core.tracking.schemas.tracking_info_run import RunInfo, ScheduledRunInfo
from dbnd._core.utils.dotdict import _as_dotted_dict
from dbnd._vendor.marshmallow import fields, post_load
from dbnd._vendor.marshmallow_enum import EnumField
from dbnd.api.serialization.common import (
    ErrorInfoSchema,
    MetricSchema,
    TargetInfoSchema,
)
from dbnd.api.serialization.run import RunInfoSchema, ScheduledRunInfoSchema
from dbnd.api.serialization.task import TaskDefinitionInfoSchema, TaskRunInfoSchema
from dbnd.api.serialization.task_run_env import TaskRunEnvInfoSchema


if typing.TYPE_CHECKING:
    from typing import Optional, List, Dict, Tuple, Sequence
    from dbnd._core.tracking.schemas.tracking_info_objects import (
        ErrorInfo,
        TaskRunEnvInfo,
        TargetInfo,
        TaskRunInfo,
        TaskDefinitionInfo,
    )


class AirflowTaskContextSchema(ApiObjectSchema):
    dag_id = fields.String()
    execution_date = fields.String()
    task_id = fields.String()
    try_number = fields.Integer(allow_none=True)
    airflow_instance_uid = fields.UUID()

    @post_load
    def make_run_info(self, data, **kwargs):
        return AirflowTaskContext(**data)


class TaskRunsInfoSchema(ApiObjectSchema):
    run_uid = fields.UUID()
    root_run_uid = fields.UUID()
    task_run_env_uid = fields.UUID()

    parent_child_map = fields.List(fields.List(fields.UUID()))
    task_runs = fields.Nested(
        TaskRunInfoSchema, many=True, exclude=("task_signature_source",)
    )
    upstreams_map = fields.List(fields.List(fields.UUID()))
    dynamic_task_run_update = fields.Boolean()

    targets = fields.Nested(TargetInfoSchema, many=True)
    task_definitions = fields.Nested(TaskDefinitionInfoSchema, many=True)
    af_context = fields.Nested(AirflowTaskContextSchema, allow_none=True)

    parent_task_run_uid = fields.UUID(allow_none=True)
    parent_task_run_attempt_uid = fields.UUID(allow_none=True)

    @post_load
    def make_run_info(self, data, **kwargs):
        return _as_dotted_dict(**data)


@attr.s
class InitRunArgs(object):
    root_run_uid = attr.ib()  # type: UUID
    run_uid = attr.ib()  # type: UUID
    task_runs_info = attr.ib()  # type: TaskRunsInfo
    driver_task_uid = attr.ib(default=None)  # type: UUID

    task_run_env = attr.ib(default=None)  # type: TaskRunEnvInfo

    new_run_info = attr.ib(
        default=None
    )  # type: Optional[RunInfo]  # we create it only for the new runs
    scheduled_run_info = attr.ib(default=None)  # type: Optional[ScheduledRunInfo]
    update_existing = attr.ib(default=False)  # type: bool
    source = attr.ib(default=UpdateSource.dbnd)  # type: UpdateSource
    af_context = attr.ib(default=None)  # type: Optional[AirflowTaskContext]

    def asdict(self):
        return attr.asdict(self, recurse=False)


class InitRunArgsSchema(ApiObjectSchema):
    run_uid = fields.UUID()
    root_run_uid = fields.UUID()

    driver_task_uid = fields.UUID(allow_none=True)

    task_run_env = fields.Nested(TaskRunEnvInfoSchema)
    task_runs_info = fields.Nested(TaskRunsInfoSchema)

    new_run_info = fields.Nested(RunInfoSchema, allow_none=True)
    scheduled_run_info = fields.Nested(ScheduledRunInfoSchema, allow_none=True)
    update_existing = fields.Boolean()
    source = fields.Str(allow_none=True)
    af_context = fields.Nested(AirflowTaskContextSchema, allow_none=True)

    @post_load
    def make_init_run_args(self, data, **kwargs):
        return InitRunArgs(**data)


@attr.s
class TaskRunAttemptUpdateArgs(object):
    task_run_uid = attr.ib()  # type: UUID
    task_run_attempt_uid = attr.ib()  # type: UUID
    timestamp = attr.ib()  # type:  datetime.datetime
    state = attr.ib()  # type: TaskRunState
    error = attr.ib(default=None)  # type: Optional[ErrorInfo]
    attempt_number = attr.ib(default=None)  # type: int
    source = attr.ib(default=UpdateSource.dbnd)  # type: UpdateSource
    start_date = attr.ib(default=None)  # type: datetime.datetime
    external_links_dict = attr.ib(default=None)


class TaskRunAttemptUpdateArgsSchema(ApiObjectSchema):
    task_run_uid = fields.UUID()
    task_run_attempt_uid = fields.UUID()
    state = EnumField(TaskRunState, allow_none=True)
    timestamp = fields.DateTime(allow_none=True)
    error = fields.Nested(ErrorInfoSchema, allow_none=True)
    attempt_number = fields.Number(allow_none=True)
    source = fields.Str(allow_none=True)
    start_date = fields.DateTime(allow_none=True)
    external_links_dict = fields.Dict(allow_none=True)

    @post_load
    def make_object(self, data, **kwargs):
        return TaskRunAttemptUpdateArgs(**data)


class InitRunSchema(_ApiCallSchema):
    init_args = fields.Nested(InitRunArgsSchema)


init_run_schema = InitRunSchema()


class AddTaskRunsSchema(_ApiCallSchema):
    task_runs_info = fields.Nested(TaskRunsInfoSchema)
    source = EnumField(UpdateSource, allow_none=True)


add_task_runs_schema = AddTaskRunsSchema()


class SetDatabandRunStateSchema(_ApiCallSchema):
    run_uid = fields.UUID(required=True)
    state = EnumField(RunState)
    timestamp = fields.DateTime(required=False, default=None, missing=None)


set_run_state_schema = SetDatabandRunStateSchema()


class SetTaskRunReusedSchema(_ApiCallSchema):
    task_run_uid = fields.UUID(required=True)
    task_outputs_signature = fields.String(required=True)


set_task_run_reused_schema = SetTaskRunReusedSchema()


class UpdateTaskRunAttemptsSchema(_ApiCallSchema):
    task_run_attempt_updates = fields.Nested(TaskRunAttemptUpdateArgsSchema, many=True)


update_task_run_attempts_schema = UpdateTaskRunAttemptsSchema()


class SetUnfinishedTasksStateShcmea(_ApiCallSchema):
    run_uid = fields.UUID()
    state = EnumField(TaskRunState)
    timestamp = fields.DateTime()


set_unfinished_tasks_state_schema = SetUnfinishedTasksStateShcmea()


class SaveTaskRunLogSchema(_ApiCallSchema):
    task_run_attempt_uid = fields.UUID(required=True)
    log_body = fields.String(allow_none=True)
    local_log_path = fields.String(allow_none=True)


save_task_run_log_schema = SaveTaskRunLogSchema()


class SaveExternalLinksSchema(_ApiCallSchema):
    task_run_attempt_uid = fields.UUID(required=True)
    external_links_dict = fields.Dict(
        name=fields.Str(), url=fields.Str(), required=True
    )


save_external_links_schema = SaveExternalLinksSchema()


class LogMetricSchema(_ApiCallSchema):
    task_run_attempt_uid = fields.UUID(required=True)
    target_meta_uid = fields.UUID(allow_none=True)
    metric = fields.Nested(MetricSchema, allow_none=True)
    source = fields.String(allow_none=True)


log_metric_schema = LogMetricSchema()


class LogMetricsSchema(_ApiCallSchema):
    metrics_info = fields.Nested(LogMetricSchema, many=True)

    @post_load
    def make_object(self, data, **kwargs):
        data["metrics_info"] = [LogMetricArgs(**m) for m in data["metrics_info"]]


log_metrics_schema = LogMetricsSchema()


class LogArtifactSchema(_ApiCallSchema):
    task_run_attempt_uid = fields.UUID(required=True)
    name = fields.String()
    path = fields.String()


log_artifact_schema = LogArtifactSchema()


@attr.s
class LogMetricArgs(object):
    task_run_attempt_uid = attr.ib()  # type: UUID
    metric = attr.ib()  # type: Metric
    source = attr.ib(default=None)  # type: Optional[str]
    target_meta_uid = attr.ib(default=None)  # type: Optional[UUID]


@attr.s
class LogTargetArgs(object):
    run_uid = attr.ib()  # type: UUID
    task_run_uid = attr.ib()  # type: UUID
    task_run_name = attr.ib()  # type: str
    task_run_attempt_uid = attr.ib()  # type: UUID
    target_path = attr.ib()  # type: str
    operation_type = attr.ib()  # type: DbndTargetOperationType
    operation_status = attr.ib()  # type: DbndTargetOperationStatus

    value_preview = attr.ib()  # type: str
    data_dimensions = attr.ib()  # type: Sequence[int]
    data_schema = attr.ib()  # type: str
    data_hash = attr.ib()  # type: str

    param_name = attr.ib(default=None)  # type: Optional[str]
    task_def_uid = attr.ib(default=None)  # type: Optional[UUID]

    def asdict(self):
        return attr.asdict(self, recurse=False)


@attr.s
class LogDataframeHistogramsArgs(object):
    run_uid = attr.ib()  # type: UUID
    task_run_uid = attr.ib()  # type: UUID
    task_run_name = attr.ib()  # type: str
    task_run_attempt_uid = attr.ib()  # type: UUID
    target_path = attr.ib()  # type: str
    operation_type = attr.ib()  # type: DbndTargetOperationType
    operation_status = attr.ib()  # type: DbndTargetOperationStatus

    value_preview = attr.ib()  # type: str
    data_dimensions = attr.ib()  # type: Sequence[int]
    data_schema = attr.ib()  # type: str
    data_hash = attr.ib()  # type: str

    descriptive_stats = attr.ib()  # type: Dict[str, Dict]
    histograms = attr.ib()  # type: Dict[str, Tuple]
    timestamp = attr.ib()  # type: datetime.datetime

    param_name = attr.ib(default=None)  # type: Optional[str]
    task_def_uid = attr.ib(default=None)  # type: Optional[UUID]


class LogTargetSchema(_ApiCallSchema):
    run_uid = fields.UUID(required=True)
    task_run_uid = fields.UUID(required=True)
    task_run_name = fields.String()
    task_run_attempt_uid = fields.UUID(required=True)

    target_path = fields.String()
    param_name = fields.String(allow_none=True)
    task_def_uid = fields.UUID(allow_none=True)
    operation_type = EnumField(DbndTargetOperationType)
    operation_status = EnumField(DbndTargetOperationStatus)

    value_preview = fields.String(allow_none=True)
    data_dimensions = fields.List(fields.Integer(), allow_none=True)
    data_schema = fields.String(allow_none=True)
    data_hash = fields.String(allow_none=True)


class LogTargetsSchema(_ApiCallSchema):
    targets_info = fields.Nested(LogTargetSchema, many=True)


log_targets_schema = LogTargetsSchema()


class HeartbeatSchema(_ApiCallSchema):
    run_uid = fields.UUID()


heartbeat_schema = HeartbeatSchema()


class AirflowTaskInfoSchema(_ApiCallSchema):
    execution_date = fields.DateTime()
    last_sync = fields.DateTime(allow_none=True)
    dag_id = fields.String()
    task_id = fields.String()
    task_run_attempt_uid = fields.UUID()
    retry_number = fields.Integer(required=False, allow_none=True)

    @post_load
    def make_object(self, data, **kwargs):
        return AirflowTaskInfo(**data)


class AirflowTaskInfosSchema(_ApiCallSchema):
    airflow_task_infos = fields.Nested(AirflowTaskInfoSchema, many=True)
    source = fields.String()
    base_url = fields.String()


airflow_task_infos_schema = AirflowTaskInfosSchema()


@attr.s
class TaskRunsInfo(object):
    run_uid = attr.ib()  # type: UUID
    root_run_uid = attr.ib()  # type: UUID

    task_run_env_uid = attr.ib()  # type: UUID

    task_runs = attr.ib(default=None)  # type: List[TaskRunInfo]
    task_definitions = attr.ib(default=None)  # type: List[TaskDefinitionInfo]
    targets = attr.ib(default=None)  # type: List[TargetInfo]

    parent_child_map = attr.ib(default=None)
    upstreams_map = attr.ib(default=None)
    dynamic_task_run_update = attr.ib(default=False)
    af_context = attr.ib(default=None)  # type: Optional[AirflowTaskContext]

    parent_task_run_uid = attr.ib(default=None)  # type: Optional[UUID]
    parent_task_run_attempt_uid = attr.ib(default=None)  # type: Optional[UUID]


@attr.s
class ScheduledJobInfo(object):
    uid = attr.ib()  # type: UUID
    name = attr.ib()  # type: str
    cmd = attr.ib()  # type: str
    start_date = attr.ib()  # type: datetime.datetime
    create_user = attr.ib()  # type: str
    create_time = attr.ib()  # type: datetime.datetime
    end_date = attr.ib(default=None)  # type: Optional[datetime.datetime]
    schedule_interval = attr.ib(default=None)  # type: Optional[str]
    catchup = attr.ib(default=None)  # type: Optional[bool]
    depends_on_past = attr.ib(default=None)  # type: Optional[bool]
    retries = attr.ib(default=None)  # type: Optional[int]
    active = attr.ib(default=None)  # type: Optional[bool]
    update_user = attr.ib(default=None)  # type: Optional[str]
    update_time = attr.ib(default=None)  # type: Optional[datetime.datetime]
    from_file = attr.ib(default=False)  # type: bool
    deleted_from_file = attr.ib(default=False)  # type: bool
    list_order = attr.ib(default=None)  # type: Optional[List[int]]
    job_name = attr.ib(default=None)  # type: Optional[str]


class ScheduledJobInfoSchema(ApiObjectSchema):
    uid = fields.UUID()
    name = fields.String()
    cmd = fields.String()
    start_date = fields.DateTime()
    create_user = fields.String()
    create_time = fields.DateTime()
    end_date = fields.DateTime(allow_none=True)
    schedule_interval = fields.String(allow_none=True)
    catchup = fields.Boolean(allow_none=True)
    depends_on_past = fields.Boolean(allow_none=True)
    retries = fields.Integer(allow_none=True)
    active = fields.Boolean(allow_none=True)
    update_user = fields.String(allow_none=True)
    update_time = fields.DateTime(allow_none=True)
    from_file = fields.Boolean(allow_none=True)
    deleted_from_file = fields.Boolean(allow_none=True)
    list_order = fields.Integer(allow_none=True)
    job_name = fields.String(allow_none=True)

    @post_load
    def make_object(self, data, **kwargs):
        return ScheduledJobInfo(**data)


class ScheduledJobArgsSchema(_ApiCallSchema):
    scheduled_job_args = fields.Nested(ScheduledJobInfoSchema)
    update_existing = fields.Boolean(default=False)


scheduled_job_args_schema = ScheduledJobArgsSchema()


@attr.s
class AirflowTaskInfo(object):
    execution_date = attr.ib()  # type: datetime.datetime
    dag_id = attr.ib()  # type: str
    task_id = attr.ib()  # type: str
    task_run_attempt_uid = attr.ib()  # type: UUID
    last_sync = attr.ib(default=None)  # type: Optional[datetime.datetime]
    retry_number = attr.ib(default=None)  # type: Optional[int]

    def asdict(self):
        return dict(
            execution_date=self.execution_date,
            last_sync=self.last_sync,
            dag_id=self.dag_id,
            task_id=self.task_id,
            task_run_attempt_uid=self.task_run_attempt_uid,
            retry_number=self.retry_number,
        )


class AirflowMonitorDataSchema(ApiObjectSchema):
    airflow_export_data = fields.Str()
    airflow_base_url = fields.Str()
    last_sync_time = fields.DateTime()


airflow_monitor_data_schema = AirflowMonitorDataSchema()



import logging
import os
import shlex
import subprocess
import sys

from subprocess import list2cmdline

from dbnd._core.configuration.environ_config import ENV_DBND_HOME
from dbnd._core.current import dbnd_context
from dbnd._core.run.databand_run import new_databand_run
from dbnd._core.task_build.task_registry import get_task_registry
from dbnd._core.tools.jupyter.notebook import notebook_run
from dbnd._core.utils import seven
from dbnd._core.utils.basics import fast_subprocess
from dbnd._core.utils.platform import windows_compatible_mode
from dbnd._core.utils.project.project_fs import abs_join


logger = logging.getLogger(__name__)


def get_environ_without_dbnd_and_airflow_vars():
    env = os.environ.copy()
    for key in list(env.keys()):
        if key.startswith("DBND") or key.startswith("AIRFLOW"):
            del env[key]
    return env


def run_dbnd_subprocess(args, retcode=255, blocking=True, **kwargs):
    # implement runner with https://docs.pytest.org/en/latest/capture.html
    # do not run in subprocess
    # main.main(['run', '--module', str(factories.__name__), ] + args)
    # return
    kwargs = kwargs.copy()
    cmd_args = list(map(str, args))
    env = kwargs.pop("env", os.environ).copy()

    # env['PYTHONUNBUFFERED'] = 'false'
    # env['PYTHONPATH'] = env.get('PYTHONPATH', '') + ':.:test'

    cmd_line = list2cmdline(cmd_args)

    logger.info(
        "Running at %s: %s", kwargs.get("cwd", "current dir"), cmd_line
    )  # To simplify rerunning failing tests

    if blocking:
        try:
            output = fast_subprocess.check_output(
                cmd_args, stderr=subprocess.STDOUT, env=env, **kwargs
            )
            # we don't decode ascii //.decode("ascii")
            output = output.decode("utf-8")
            logger.info("Cmd line %s output:\n %s", cmd_line, output)
            return output
        except subprocess.CalledProcessError as ex:
            logger.error(
                "Failed to run %s :\n\n\n -= Output =-\n%s\n\n\n -= See output above =-",
                cmd_line,
                ex.output.decode("utf-8", errors="ignore"),
            )
            if ex.returncode == retcode:
                return ex.output.decode("utf-8")
            raise ex
    else:
        return subprocess.Popen(cmd_args, stderr=subprocess.STDOUT, env=env, **kwargs)


def run_test_notebook(notebook):
    return notebook_run(notebook)


def run_subprocess__airflow(args, retcode=255, **kwargs):
    return run_dbnd_subprocess(
        args=["dbnd-airflow"] + args,
        cwd=kwargs.pop("cwd", os.environ[ENV_DBND_HOME]),
        retcode=retcode,
        **kwargs
    )


def run_dbnd_subprocess__dbnd(args, retcode=255, **kwargs):
    if isinstance(args, str):
        args = shlex.split(args, posix=not windows_compatible_mode)
    return run_dbnd_subprocess(
        args=[sys.executable, "-m", "dbnd"] + args,
        cwd=kwargs.pop("cwd", os.environ[ENV_DBND_HOME]),
        retcode=retcode,
        **kwargs
    )


def run_dbnd_subprocess__with_home(args, retcode=255, **kwargs):
    return run_dbnd_subprocess(
        args=[sys.executable] + args,
        cwd=kwargs.pop("cwd", os.environ[ENV_DBND_HOME]),
        retcode=retcode,
        **kwargs
    )


def run_dbnd_subprocess__dbnd_run(args, module=None, retcode=255, **kwargs):
    if module:
        args = ["--module", str(module.__name__)] + args

    return run_dbnd_subprocess__dbnd(["run"] + args, retcode=retcode, **kwargs)


def build_task(root_task, **kwargs):
    from dbnd import new_dbnd_context

    with new_dbnd_context(conf={root_task: kwargs}):
        return get_task_registry().build_dbnd_task(task_name=root_task)


@seven.contextlib.contextmanager
def initialized_run(task_or_task_name):
    with new_databand_run(
        context=dbnd_context(), task_or_task_name=task_or_task_name
    ) as r:
        r._init_without_run()
        yield r


def dbnd_module_path():
    return abs_join(__file__, "..", "..", "..", "..")


def dbnd_examples_path():
    return abs_join(dbnd_module_path(), "..", "..", "examples", "src")


def import_all_modules(src_dir, package, excluded=None):
    packagedir = os.path.join(src_dir, package)
    errors = []
    imported_modules = []

    def import_module(p):
        try:
            logger.info("Importing %s", p)
            __import__(p)
            imported_modules.append(p)
        except Exception as ex:
            errors.append(ex)
            logger.exception("Failed to import %s", p)

    excluded = excluded or set()

    for root, subdirs, files in os.walk(packagedir):
        package = os.path.relpath(root, start=src_dir).replace(os.path.sep, ".")

        if any([p in root for p in excluded]):
            continue

        if "__init__.py" not in files:
            continue

        import_module(package)

        for f in files:
            if f.endswith(".py") and not f.startswith("_"):
                import_module(package + "." + f[:-3])

    return imported_modules



import logging
import os
import sys

from contextlib import contextmanager
from logging.config import dictConfig
from typing import Optional

from dbnd._core.configuration.dbnd_config import config as dbnd_config
from dbnd._core.log.logging_utils import setup_log_file


END_OF_LOG_MARK = "end_of_log"
logger = logging.getLogger(__name__)


FORMAT_FULL = "[%(asctime)s] {%(filename)s:%(lineno)d} %(levelname)s - %(message)s"
FORMAT_SIMPLE = "%(asctime)s %(levelname)s - %(message)s"
FORMAT_COLORLOG = "[%(asctime)s] %(log_color)s%(levelname)s %(reset)s - %(message)s"


def basic_logging_config(
    filename=None,
    log_level=logging.INFO,
    console_stream=sys.stderr,
    console_formatter_name="formatter_colorlog",
    file_formatter_name="formatter_full",
):
    # type: (...) -> Optional[dict]

    config = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "formatter_full": {"format": FORMAT_FULL},
            "formatter_simple": {"format": FORMAT_SIMPLE},
            "formatter_colorlog": {
                "()": "dbnd._vendor.colorlog.ColoredFormatter",
                "format": FORMAT_COLORLOG,
                "reset": True,
            },
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "stream": console_stream,
                "formatter": console_formatter_name,
            }
        },
        "root": {"handlers": ["console"], "level": log_level},
    }
    if filename:
        setup_log_file(filename)
        config["handlers"]["file"] = {
            "class": "logging.FileHandler",
            "formatter": file_formatter_name,
            "filename": filename,
            "encoding": "utf-8",
        }
        config["root"]["handlers"].append("file")

    return config


def get_sentry_logging_config(sentry_url, sentry_env):
    import raven.breadcrumbs

    for ignore in (
        "sqlalchemy.orm.path_registry",
        "sqlalchemy.pool.NullPool",
        "raven.base.Client",
    ):
        raven.breadcrumbs.ignore_logger(ignore)

    return {
        "exception": {
            "level": "ERROR",
            "class": "raven.handlers.logging.SentryHandler",
            "dsn": sentry_url,
            "environment": sentry_env,
        }
    }


def configure_logging_dictConfig(dict_config):
    try:
        dictConfig(dict_config)
    except Exception:
        logging.exception("Unable to configure logging using %s!", dict_config)
        raise


def configure_basic_logging(log_file):
    """
    Simple databand logging, called from main and other commands
    """
    configure_logging_dictConfig(basic_logging_config(filename=log_file))



from dbnd._core.tracking.backends.channels.abstract_channel import TrackingChannel
from dbnd.api.tracking_api import (
    add_task_runs_schema,
    airflow_monitor_data_schema,
    airflow_task_infos_schema,
    heartbeat_schema,
    init_run_schema,
    log_artifact_schema,
    log_metrics_schema,
    log_targets_schema,
    save_external_links_schema,
    save_task_run_log_schema,
    scheduled_job_args_schema,
    set_run_state_schema,
    set_task_run_reused_schema,
    set_unfinished_tasks_state_schema,
    update_task_run_attempts_schema,
)


class MarshmallowMixin:
    SCHEMA_BY_HANDLER_NAME = {
        TrackingChannel.init_scheduled_job.__name__: scheduled_job_args_schema,
        TrackingChannel.init_run.__name__: init_run_schema,
        TrackingChannel.add_task_runs.__name__: add_task_runs_schema,
        TrackingChannel.set_run_state.__name__: set_run_state_schema,
        TrackingChannel.set_task_reused.__name__: set_task_run_reused_schema,
        TrackingChannel.update_task_run_attempts.__name__: update_task_run_attempts_schema,
        TrackingChannel.set_unfinished_tasks_state.__name__: set_unfinished_tasks_state_schema,
        TrackingChannel.save_task_run_log.__name__: save_task_run_log_schema,
        TrackingChannel.save_external_links.__name__: save_external_links_schema,
        TrackingChannel.log_targets.__name__: log_targets_schema,
        TrackingChannel.log_metrics.__name__: log_metrics_schema,
        TrackingChannel.log_artifact.__name__: log_artifact_schema,
        TrackingChannel.heartbeat.__name__: heartbeat_schema,
        TrackingChannel.save_airflow_task_infos.__name__: airflow_task_infos_schema,
        TrackingChannel.save_airflow_monitor_data.__name__: airflow_monitor_data_schema,
    }

    def get_schema_by_handler_name(self, handler_name):
        return self.SCHEMA_BY_HANDLER_NAME[handler_name]



import abc
import logging

from dbnd._core.errors import friendly_error
from dbnd._core.utils.basics.nothing import NOTHING, is_not_defined
from targets import Target
from targets.caching import TARGET_CACHE, TargetCacheKey
from targets.errors import NotADirectory
from targets.extras.file_ctrl import ObjectMarshallingCtrl
from targets.marshalling import get_marshaller_ctrl
from targets.utils.performance import target_timeit_log


logger = logging.getLogger(__name__)


class DataTarget(Target):
    """
    Target with Data
    """

    target_no_traverse = True

    def __init__(self, properties=None, source=None):
        """
        Initializes a DataTarget instance.

        :param str path: the path associated with this DataTarget.
        """
        super(DataTarget, self).__init__(properties=properties, source=source)

        try:
            import pandas
            from targets.extras.pandas_ctrl import PandasMarshallingCtrl

            self.as_pandas = PandasMarshallingCtrl(self)
        except ImportError:
            pass

        self.as_object = ObjectMarshallingCtrl(self)

    @abc.abstractmethod
    def open(self, mode="r"):
        """
        Open the DataTarget target.

        This method returns a file-like object which can either be read from or written to depending
        on the specified mode.

        :param str mode: the mode `r` opens the DataTarget in read-only mode, whereas `w` will
                         open the DataTarget in write mode. Subclasses can implement
                         additional options.
        """
        pass

    def _dt(self):
        from targets.data_transaction import DataTransaction

        return DataTransaction(self)

    @abc.abstractmethod
    def remove(self):
        """
        Remove the resource at the path specified by this DataTarget.

        This method is implemented by using :py:attr:`fs`.
        """
        pass

    def mark_success(self):
        """
        marks current target as exists.
        :return:
        """
        pass

    def list_partitions(self):
        return [self]

    def partition(self, name=NOTHING, **kwargs):
        raise NotADirectory(
            "You can create paritions for DirTarget only. "
            "Please validate that you have .folder flag and you path ends with slash: %s"
            % str(self)
        )

    def file(self, name=NOTHING, extension=NOTHING, config=NOTHING, **kwargs):
        """
        :param config:
        :param name: file name of the partition. if not provided - "part-%04d" % ID
        :param extension: extension. if not provided -> default extension will be used
        :return: FileTarget: that represents the partition.
        """
        config = self.config if is_not_defined(config) else config
        return self.partition(
            name=name, extension=extension, config=config.as_file(), **kwargs
        )

    def folder(self, name=NOTHING, extension=NOTHING, config=NOTHING, **kwargs):
        """
        :param config:
        :param name: file name of the partition. if not provided - "part-%04d" % ID
        :param extension: extension. if not provided -> default extension will be used
        :return: DirTarget: that represents the partition.
        """

        config = self.config if is_not_defined(config) else config
        return self.partition(
            name=name, extension=extension, config=config.as_folder(), **kwargs
        )

    def save(self, value):
        return self.dump(value=value)

    def dump(self, value, value_type=None, **kwargs):
        from targets.values import get_value_type_of_obj, get_value_type_of_type
        from targets.values import ObjectValueType

        if value_type:
            value_type = get_value_type_of_type(value_type)
        else:
            value_type = get_value_type_of_obj(value, ObjectValueType())
        try:
            m = get_marshaller_ctrl(self, value_type=value_type)

            with target_timeit_log(self, "marshalling"):
                m.dump(value, **kwargs)
        except Exception as ex:
            raise friendly_error.failed_to_write_task_output(
                ex, self, value_type=value_type
            )
        cache_key = TargetCacheKey(target=self, value_type=value_type)
        TARGET_CACHE[cache_key] = value

    def load(self, value_type, **kwargs):
        cache_key = TargetCacheKey(target=self, value_type=value_type)
        if cache_key in TARGET_CACHE:
            logger.info("Using cached data value for target='%s'", self)
            return TARGET_CACHE[cache_key]

        m = get_marshaller_ctrl(self, value_type)

        with target_timeit_log(self, "unmarshalling"):
            value = m.load(**kwargs)

        TARGET_CACHE[cache_key] = value

        return value

    def touch(self):
        return self.as_object.touch()

    # aliases fro the most used functions
    def readlines(self, **kwargs):
        return self.as_object.readlines(**kwargs)

    def read(self, **kwargs):
        return self.as_object.read(**kwargs)

    def read_obj(self, **kwargs):
        return self.as_object.read_obj(**kwargs)

    def read_pickle(self, **kwargs):
        return self.as_object.read_pickle(**kwargs)

    def write(self, fn, **kwargs):
        self.as_object.write(fn, **kwargs)

    def writelines(self, fn, **kwargs):
        self.as_object.writelines(fn, **kwargs)

    def write_pickle(self, obj, **kwargs):
        self.as_object.write_pickle(obj, **kwargs)

    def write_numpy_array(self, arr, **kwargs):
        self.as_object.write_numpy_array(arr, **kwargs)

    def read_df(self, **kwargs):
        return self.as_pandas.read(**kwargs)

    def read_df_partitioned(self, **kwargs):
        for t in self.as_pandas.read_partitioned(**kwargs):
            yield t

    def write_df(self, df, **kwargs):
        return self.as_pandas.to(df, **kwargs)



import logging
import typing

from datetime import timedelta
from math import floor, log10

import six

from dbnd._core.constants import TaskRunState, UpdateSource
from dbnd._core.current import is_verbose
from dbnd._core.tracking.backends import TrackingStore
from dbnd._core.tracking.schemas.metrics import Metric
from dbnd._core.utils.timezone import utcnow
from dbnd._vendor.ascii_graph import Pyasciigraph


if typing.TYPE_CHECKING:
    from typing import List
    from dbnd._core.task_run.task_run import TaskRun
logger = logging.getLogger(__name__)


class ConsoleStore(TrackingStore):
    def __init__(self):
        super(ConsoleStore, self).__init__()
        self.max_log_value_len = 50
        self.verbose = is_verbose()
        self.ascii_graph = Pyasciigraph()

    def init_scheduled_job(self, scheduled_job, update_existing):
        super(ConsoleStore, self).init_scheduled_job(scheduled_job, update_existing)

    def init_run(self, run):
        if run.is_orchestration:
            logger.info(
                run.describe.run_banner(
                    "Running Databand!", color="cyan", show_run_info=True
                )
            )

            if run.context.name == "interactive":
                from dbnd._core.tools import ipython

                ipython.show_run_url(run.run_url)

    def set_task_reused(self, task_run):
        task = task_run.task
        logger.info(
            task.ctrl.visualiser.banner(
                "Task %s has been completed already!" % task.task_id,
                "magenta",
                task_run=task_run,
            )
        )

    def set_task_run_state(self, task_run, state, error=None, timestamp=None):
        super(ConsoleStore, self).set_task_run_state(task_run=task_run, state=state)
        task = task_run.task

        # optimize, don't print success banner for fast running tasks
        start_time = task_run.start_time or utcnow()
        quick_task = task_run.finished_time and (
            task_run.finished_time - start_time
        ) < timedelta(seconds=5)

        show_simple_log = not self.verbose and (
            task_run.task.task_is_system or quick_task
        )

        level = logging.INFO
        color = "cyan"
        task_friendly_id = task_run.task_af_id
        if state in [TaskRunState.RUNNING, TaskRunState.QUEUED]:
            task_msg = "Running task %s" % task_friendly_id

        elif state == TaskRunState.SUCCESS:
            task_msg = "Task %s has been completed!" % task_friendly_id
            color = "green"

        elif state == TaskRunState.FAILED:
            task_msg = "Task %s has failed!" % task_friendly_id
            color = "red"
            level = logging.ERROR
            if task_run.task.get_task_family() != "_DbndDriverTask":
                show_simple_log = False

        elif state == TaskRunState.CANCELLED:
            task_msg = "Task %s has been canceled!" % task_friendly_id
            color = "red"
            level = logging.ERROR

        else:
            task_msg = "Task %s at %s state" % (task_friendly_id, state)

        if show_simple_log:
            logger.log(level, task_msg)
            return

        try:
            logger.log(
                level, task.ctrl.visualiser.banner(task_msg, color, task_run=task_run)
            )
        except Exception as ex:
            logger.log(level, "%s \nfailed to create banner: %s" % (task_msg, ex))

    def add_task_runs(self, run, task_runs):
        pass

    def save_external_links(self, task_run, external_links_dict):
        task = task_run.task
        attempt_uid = task_run.task_run_attempt_uid

        logger.info(
            "%s %s has URLs to external resources: \n\t%s",
            task,
            attempt_uid,
            "\t\n".join(
                "%s: %s" % (k, v) for k, v in six.iteritems(external_links_dict)
            ),
        )

    def is_ready(self):
        return True

    def log_histograms(self, task_run, key, value_meta, timestamp):
        for hist_name, hist_values in value_meta.histograms.items():
            for line in self.ascii_graph.graph(
                label="Histogram logged: {}.{}".format(key, hist_name),
                data=list(zip(*reversed(hist_values))),
            ):
                logger.info(line)
        # TODO: Add more compact logging if user opts out

        if not self.verbose:
            return

        for key, value in value_meta.histogram_system_metrics.items():
            if isinstance(value, float):
                value = round(value, 3)
            logger.info("%s: %s", key, value)

    @staticmethod
    def _format_value(value):
        """
        Convert value to string.
        Additionally forces float values to NOT use scientific notation.
        """
        if isinstance(value, float):
            value_width = abs(floor(log10(value)))
            # default python precision threshold to switch to scientific representation of floats
            if value > 1 or value_width < 4:
                return str(value)
            return "{:{width}.{prec}f}".format(
                value, width=value_width, prec=value_width + 4
            )
        return str(value)

    def log_metrics(self, task_run, metrics):
        # type: (TaskRun, List[Metric]) -> None
        if len(metrics) == 1:
            m = metrics[0]
            logger.info(
                "Metric logged: {}={}".format(m.key, self._format_value(m.value))
            )
        else:
            metrics_str = "\n\t".join(
                ["{}={}".format(m.key, self._format_value(m.value)) for m in metrics]
            )
            logger.info("Metrics logged:\n\t%s", metrics_str)



from dbnd._core.errors import DatabandRuntimeError


def data_frame_dict_value_should_be_dict(target, value):
    return DatabandRuntimeError(
        "Can't dump %s to %s': value should be Mapping[str,DataFrame]"
        % (type(value), target)
    )


def dump_to_multi_target(target, value):
    return DatabandRuntimeError(
        "Can't dump %s to %s': target is multi target (read only)"
        % (type(value), target)
    )


def dump_generator_to_file(target):
    return DatabandRuntimeError(
        "Can't dump generator (multiple parts) to %s': target is not directory" % target
    )


def no_marshaller(target, config, value_type, options_message):
    return DatabandRuntimeError(
        "There is no defined way to read/write value of type '{type}' with file format '{format}'. ".format(
            type=value_type.type, format=config.format
        ),
        help_msg="You can provide the expected format of {target}"
        " using --PARAMETER-target switch, for example --my-input--target csv.  "
        "{options_message}".format(target=target, options_message=options_message),
    )


def no_format(target, options_message):
    return DatabandRuntimeError(
        "Can't read file unknown format. "
        "It's file extension can not be mapped to any of known formats.".format(),
        help_msg="You can "
        "provide the expected format of {target} using --PARAMETER--target switch. "
        "Otherwise you can register new extension via 'register_format/register_compression' or "
        "{options_message} For example: --my-input--target csv.gz".format(
            target=target, options_message=options_message
        ),
    )


def type_without_parse_from_str(value_type):
    return DatabandRuntimeError(
        "Value type {} doesn't support parse_from_str".format(value_type),
        help_msg="Review type implementation and set support_parse_from_str flag to True",
    )


def failed_to_save_value__wrong_type(value, target, expected_type):
    return DatabandRuntimeError(
        "Can not save value at {target}, "
        "expected type is '{expected_type}', got '{value_type}'".format(
            target=target, expected_type=expected_type, value_type=type(value)
        ),
        help_msg="Review type implementation and set support_parse_from_str flag to True",
    )


def target_must_be_local_for_tensorflow_marshalling(target):
    return DatabandRuntimeError(
        "Can not read value of tensorflow model in path {path}! Path must be local!".format(
            path=target.path
        ),
        help_msg="To marshall tensorflow objects you must use 'require_local_access`, e.g.:\n@task("
        "result=output.tfmodel.require_local_access[tf.keras.models.Model])\ndef my_task(p1, p2):...",
    )



import contextlib
import logging
import os
import signal
import subprocess
import sys

from time import sleep, time

from dbnd._core.constants import RunState
from dbnd._core.tracking.backends import TrackingStore
from dbnd._core.utils.basics.format_exception import format_exception_as_str
from dbnd._vendor.psutil.vendorized_psutil import pid_exists


logger = logging.getLogger(__name__)

TERMINATE_WAIT_TIMEOUT = 5


@contextlib.contextmanager
def start_heartbeat_sender(task_run):
    run = task_run.run
    settings = run.context.settings
    core = settings.core
    heartbeat_interval_s = settings.run.heartbeat_interval_s

    if heartbeat_interval_s > 0:
        sp = None
        heartbeat_log_fp = None
        try:
            try:
                cmd = [
                    sys.executable,
                    "-m",
                    "dbnd",
                    "send-heartbeat",
                    "--run-uid",
                    str(run.run_uid),
                    "--driver-pid",
                    str(os.getpid()),
                    "--heartbeat-interval",
                    str(heartbeat_interval_s),
                    "--tracker",
                    ",".join(core.tracker),
                    "--tracker-api",
                    core.tracker_api,
                ]
                if core.databand_url:
                    cmd += ["--databand-url", core.databand_url]

                if settings.run.heartbeat_sender_log_to_file:
                    heartbeat_log_file = task_run.log.local_heartbeat_log_file
                    heartbeat_log_fp = heartbeat_log_file.open("w")
                    stdout = heartbeat_log_fp
                    logger.info(
                        "Starting heartbeat with log at %s using cmd: %s",
                        heartbeat_log_file,
                        subprocess.list2cmdline(cmd),
                    )
                else:
                    stdout = None
                    logger.info(
                        "Starting heartbeat using cmd: %s", subprocess.list2cmdline(cmd)
                    )

                sp = subprocess.Popen(cmd, stdout=stdout, stderr=subprocess.STDOUT)
            except Exception as ex:
                logger.info(
                    "Failed to spawn heartbeat process, you can disable it via [task]heartbeat_interval_s=0  .\n %s",
                    ex,
                )
                raise ex
            yield
        finally:
            if sp:
                sp.terminate()

                try:
                    sp.wait(timeout=TERMINATE_WAIT_TIMEOUT)
                except Exception:
                    logger.warning(
                        "waited %s seconds for the heartbeat sender to exit but it still hasn't exited",
                        TERMINATE_WAIT_TIMEOUT,
                    )

            if heartbeat_log_fp:
                heartbeat_log_fp.close()
    else:
        logger.info(
            "run heartbeat sender disabled (set task.heartbeat_interval_s to value > 0)"
        )
        yield


def send_heartbeat_continuously(
    run_uid, tracking_store, heartbeat_interval_s, driver_pid
):  # type: (str, TrackingStore, int, int) -> None
    logger.info(
        "[heartbeat sender] starting heartbeat sender process (pid %s) with a send interval of %s seconds"
        % (os.getpid(), heartbeat_interval_s)
    )

    try:
        while True:
            loop_start = time()
            try:
                if not pid_exists(
                    driver_pid
                ):  # failsafe, in case the driver process died violently
                    logger.info(
                        "[heartbeat sender] driver process %s stopped, stopping heartbeat sender",
                        driver_pid,
                    )
                    return

                run_state = tracking_store.heartbeat(run_uid=run_uid)
                logger.debug("[heartbeat sender] sent heartbeat")
                if run_state == RunState.SHUTDOWN.value:
                    logger.info(
                        "[heartbeat sender] received run state SHUTDOWN: killing driver process"
                    )
                    os.kill(driver_pid, signal.SIGTERM)
            except KeyboardInterrupt:
                logger.info(
                    "[heartbeat sender] stopping heartbeat sender process due to interrupt"
                )
                return
            except Exception:
                logger.error(
                    "[heartbeat sender] failed to send heartbeat: %s",
                    format_exception_as_str(),
                )

            time_to_sleep_s = max(0, time() + heartbeat_interval_s - loop_start)
            if time_to_sleep_s > 0:
                sleep(time_to_sleep_s)
    except KeyboardInterrupt:
        return
    except Exception:
        logger.exception("[heartbeat sender] Failed to run heartbeat")
    finally:
        logger.info("[heartbeat sender] stopping heartbeat sender")
        sys.exit(0)
